"""
##############################################################################

File containing a class intend to facilitate clustering analysis.
The functions are based on the package scikit learn.

Authors: Ana Marta Sequeira

Date: 06/2019 altered 01/2021

Email:

##############################################################################
"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.cluster.hierarchy import dendrogram, linkage
from sklearn.cluster import AgglomerativeClustering
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.metrics import accuracy_score
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay

from .ml.utils import timer


class Cluster:
    """
    The cluster class aims to perform and plot clustering analysis
    Based on scikit learn.
    """

    def __init__(self, fps_x, target=None, report_name=None):
        """
        init function. When the class is called a x dataset must be provided (with features or encodings).
        Target is optional and only needed f you want to validate your cluster.
        If given a report name,  the class will automatically generate report for this analysis.
        :param fps_x: dataset X_data
        :param target: column with class labels. None by default
        :param report_name: str. if given, the class will automatically retrieve a txt report with results from
        the functions called inside class.
        """

        self.X_data = fps_x
        self.Y_data = target
        self.y_kmeans = None
        self.clf = None
        self.y_labels = None
        self.centroids = None
        self.affinity_matrix = None
        self.report_name = report_name
        if self.report_name:
            self._report(str(self.report_name))

    def _report(self, info):
        filename = str(self.report_name)
        with open(filename, 'a+') as file:
            if isinstance(info, str):
                file.writelines(info)
            else:
                for l in info:
                    file.writelines('\n{}'.format(l))

    def kmeans_table(self, rownames=('clusters',)):
        """
        Function to generate a cross table with results from Kmeans
        :param rownames: row names. ('clusters',) by default.
        :return: table
        """
        table = pd.crosstab(self.y_labels, columns=self.Y_data, rownames=rownames)
        return table

    def kmeans_seaborn_table(self, path_save='', show=True):
        """
        Function to generate a seaborn graphic with table results from kmeans
        :param path_save:  path to save file generated. By default ''
        :param show: If to display or not the graphic. True by default.
        :return: seaborn plot with k means results
        """
        plt.clf()
        mat = confusion_matrix(self.Y_data, self.y_labels)
        disp = ConfusionMatrixDisplay(confusion_matrix=mat.T)
        disp.plot()
        # sns.heatmap(mat, square=True, annot=True, fmt='d', cbar=False)
        # plt.xlabel('true label')
        # plt.ylabel('predicted label')
        # save and empty plt
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()

    def clust_evaluation(self, y_data):
        """
        Function to evaluate clustering results on target data.
        :param y_data:  list of targets correspondent to fps_x. If not given, it will retrieve the target from init.
        :return: accuracy evaluation of the kmeans clustering prediction
        """
        if self.Y_data is None:
            self.Y_data = y_data

        accuracy = accuracy_score(self.Y_data, self.y_labels)
        print('accuracy', accuracy)
        # table
        table = self.kmeans_table()
        print(table)

        # fig table
        if self.report_name is not None:
            kmeans_table_path = str(self.report_name.split('.', 1)[0] + 'kmeans_table.png')
        else:
            kmeans_table_path = None
        self.kmeans_seaborn_table(path_save=kmeans_table_path, show=True)

        things_to_report = ['accuracy: {}'.format(accuracy), table]
        return things_to_report

    @timer
    def run_kmeans(self, max_iter=300, n_clusters=None, init='k-means++', random_state=42, **params):
        """
        Function that performs K means cluster.
        :param init: init function for kmeans . 'k-means++' by default.
        :param max_iter: number of max terations for cluster (300 by default)
        :param n_clusters: if None, it will define the number of clusters as the number of existing labels
        :return: cross table with counts for labels vs classification in clusters
        """

        if not n_clusters:
            n_clusters = len(np.unique(self.Y_data))

        saved_args = locals()

        self.clf = KMeans(n_clusters=n_clusters, init=init, max_iter=max_iter, random_state=random_state, **params)
        self.clf.fit(self.X_data)
        self.y_labels = self.clf.labels_
        self.centroids = self.clf.cluster_centers_
        self.y_kmeans = self.clf.predict(self.X_data)

        if self.Y_data is not None:
            things_to_report = self.clust_evaluation(self.Y_data)
        else:
            s = 'no true labels / y is given to evaluate k means'
            things_to_report = [s]
            print(s)

        if self.report_name is not None:
            self._report([self.run_kmeans.__name__, saved_args, 'y_labels_\n{}'.format(self.y_labels),
                          'centroids_\n{}'.format(self.centroids),
                          things_to_report])

        return self.clf, self.y_labels, self.centroids

    @timer
    def run_minibatch_kmeans(self, max_iter=300, batch_size=100, n_clusters=None, init='k-means++', random_state=42,
                             **params):
        """
        Function that performs K means cluster using mini batches.
        :param init: init function for kmeans . 'k-means++' by default.
        :param max_iter: number of max terations for cluster (300 by default)
        :param n_clusters: if None, it will define the number of clusters as the number of existing labels
        :param batch_size: batch size to use in mini batch. 100 by default.
        :return: cross table with counts for labels vs classification in clusters
        """
        if not n_clusters:
            n_clusters = len(np.unique(self.Y_data))
        saved_args = locals()

        self.clf = MiniBatchKMeans(n_clusters=n_clusters, init=init, batch_size=batch_size, max_iter=max_iter,
                                   random_state=random_state, **params)
        self.clf.fit(self.X_data)
        self.y_labels = self.clf.labels_
        self.centroids = self.clf.cluster_centers_
        self.y_kmeans = self.clf.predict(self.X_data)

        if self.Y_data is not None:
            things_to_report = self.clust_evaluation(self.Y_data)
        else:
            s = 'no true labels / y is given to evaluate k means'
            things_to_report = [s]
            print(s)

        if self.report_name:
            self._report([self.run_minibatch_kmeans.__name__, saved_args, 'y_labels_\n{}'.format(self.y_labels),
                          'centroids_\n{}'.format(self.centroids),
                          things_to_report])

        return self.clf, self.y_labels, self.centroids

#######################################################################################################################################
    # multiprocessing
    # accept the other matrix before
    # put the dendogram outside this?
    @timer
    def hierarchical_dendogram(self, metric='correlation', method='complete', path_save='', show=True,
                               truncate_mode='level', p=3, **params):
        """
        Perform hierarchical clustering scipy matrices
        :param metric: distance metric to use in the case that y is a collection of observation vectors. eg. 'correlation', 'euclidean'
        see (https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.pdist.html#scipy.spatial.distance.pdist)
        :param method: method to be used. exemples: 'complete', 'single', 'average', 'ward'
        see (https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html)
        test if no labels work
        :param truncate_mode: 'level' by default.
        :return: dendogram of the clustered data The hierarchical clustering encoded as a linkage matrix.
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
        https://scipy.github.io/devdocs/generated/scipy.cluster.hierarchy.dendrogram.html
        ** params for dendogram
        """
        saved_args = locals()
        plt.clf()
        Z = linkage(self.X_data, metric=metric, method=method)

        # The hierarchical clustering encoded as a linkage matrix.
        plt.figure(figsize=(25, 10))
        plt.title('Hierarchical Clustering Dendrogram metric: {}, method: {}'.format(metric, method))
        plt.xlabel('sample index')
        plt.ylabel('distance')
        labels = list(range(len(self.X_data)))
        label_colors = None
        if self.Y_data is not None:
            labels = list(self.Y_data)
            color_labels = np.unique(self.Y_data)
            rgb_values = sns.color_palette("nipy_spectral", len(color_labels))  # 'set2'
            label_colors = dict(zip(color_labels, rgb_values))

        dendrogram(Z,
                   labels=np.array(labels),
                   truncate_mode=truncate_mode,
                   p=p,
                   leaf_rotation=90,  # rotate the x axis labels
                   leaf_font_size=8, **params)  # font size for the x axis labels
        ax = plt.gca()
        xlabels = ax.get_xmajorticklabels()
        for lbl in xlabels:
            print(lbl)
            print(label_colors)
            lbl.set_color(label_colors[lbl.get_text()])
        if path_save is not None:
            plt.savefig(fname=path_save, bbox_inches='tight', pad_inches=0)
        if show is True:
            plt.show()
        plt.clf()
        if self.report_name:
            self._report([self.hierarchical_dendogram.__name__, saved_args,
                          'y_labels_\n{}'.format(self.y_labels)])
        return Z

    # do the same as the scipy but it is taking to long. because te matrix is a little bit different
    # it needs acessory functions to plot the dendogram
    # check if functions are ok and if are worth to keep
    # check if can be combine the fuctions to the graphic from scipy matrix, as it is the same dendogram
    # and use the same function to plot
    @timer
    # check if it can receive a linkage matrix
    def run_hierarchical(self, n_clusters=None, affinity='euclidean', linkage='ward', y_data=None, **params):
        """
        Perform hierarchical clustering using scikit learn
        https://scikit-learn.org/stable/modules/clustering.html#hierarchical-clustering

        :param n_clusters: number of clusters in hierarchical clustering. If None will be set to the number of labels
        :param affinity: affinity function. default euclidean
        :param linkage: type of linkage . default ward
        :param y_data: true labels of dataset. Default None
        :param params: other parameters to agglomerative clustering function
        For more details please see scikit learn documentation
        :return: agglomerative clustering model, labels predicted and leaves of cluster
        """
        # https://stackabuse.com/hierarchical-clustering-with-python-and-scikit-learn/
        if not n_clusters:
            n_clusters = len(np.unique(self.Y_data))
        saved_args = locals()
        hmodel = AgglomerativeClustering(n_clusters=n_clusters, affinity=affinity, linkage=linkage, compute_distances=True,  **params)
        hmodel = hmodel.fit(self.X_data)
        self.y_labels = hmodel.labels_
        self.n_leaves = hmodel.n_leaves_
        self.n_components = hmodel.n_connected_components_
        self.hmodel = hmodel

        if self.Y_data or y_data is not None:
            things_to_report = self.clust_evaluation(y_data)
        else:
            s = 'no true labels / y is given to evaluate k means'
            things_to_report = [s]
            print(s)

        if self.report_name:
            self._report([self.run_hierarchical.__name__, saved_args, 'y_labels_\n{}'.format(self.y_labels),
                          'n_leaves_\n{}'.format(self.n_leaves),
                          'n_connected_components_\n{}'.format(self.n_components),
                          things_to_report])

        return hmodel, self.y_labels, self.n_leaves

    def _scikit_dendrogram(self, model, **kwargs):
        # code from scikit learn
        # https://scikit-learn.org/stable/auto_examples/cluster/plot_agglomerative_dendrogram.html
        # model = agglomerative cluster.fit(x

        # Create linkage matrix and then plot the dendrogram
        # create the counts of samples under each node

        counts = np.zeros(model.children_.shape[0])
        n_samples = len(model.labels_)
        for i, merge in enumerate(model.children_):
            current_count = 0
            for child_idx in merge:
                if child_idx < n_samples:
                    current_count += 1  # leaf node
                else:
                    current_count += counts[child_idx - n_samples]
            counts[i] = current_count

        linkage_matrix = np.column_stack([model.children_, model.distances_,
                                          counts]).astype(float)

        # Plot the corresponding dendrogram
        dendrogram(linkage_matrix, **kwargs)

    def plot_dendrogram(self, model=None, title='Hierarchical Clustering Dendrogram', truncate_mode='level', p=3,
                        path_save='', show=True):
        """
        Plot dendogram of scikit hierarchical clustering model
        :param model: scikit hierarchical clustering model
        :param title: title of the plot. 'Hierarchical Clustering Dendrogram' by default
        :param truncate_mode: Truncation is used to condense the dendrogram. By default 'level'.
        in 'level' no more than p levels of the dendrogram tree are displayed. A “level” includes all nodes with p merges from the last merge.
        'lastp' or None are also accepted.
        :param p: The p parameter for truncate_mode. By default 3
        :param path_save: If not None, path to save the dendogram
        :param show: wether to display the dendogram. True by default.
        :return:
        """
        if model is None:
            model = self.hmodel
        plt.clf()
        plt.title(title)
        # plot the top three levels of the dendrogram
        self._scikit_dendrogram(model, truncate_mode=truncate_mode, p=p)
        plt.xlabel("Number of points in node (or index of point if no parenthesis).")
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()

    # with parallel_backend('loki')
    # from sklearn.neighbors import kneighbors_graph
    # connectivity = kneighbors_graph(X, n_neighbors=10)
    # get a connectivity matrix before to do the agglomerative clustering first
    # https://scikit-learn.org/stable/modules/clustering.html#hierarchical-clustering
    # https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.kneighbors_graph.html#sklearn.neighbors.kneighbors_graph
    # https://stackoverflow.com/questions/47321133/sklearn-hierarchical-agglomerative-clustering-using-similarity-matrix

    # check the nightly version of scikit learn. useful information all the pipeline. ROC curves automatic...
    # https://scikit-learn.org/dev/common_pitfalls.html
    # https://scikit-learn.org/dev/computing/parallelism.html#parallelism
    # check for parallel runs. understand
    # https://joblib.readthedocs.io/en/latest/parallel.html#joblib.parallel_backend
    # https://scikit-learn.org/dev/computing/parallelism.html
    # check these three clusters. not scalable for large volumes of data
    # @timer nt scalable for many many features but good try / validate with other datasets.
    # https://scikit-learn.org/stable/modules/clustering.html#dbscan
    # def spectral_clustering(self, n_clusters=None, affinity='rbf',
    #                         assign_labels='kmeans', **params): #affinity = rbf   nearest_neighbors
    #     """
    #     allow kmeans to discover non linear boundaries
    #     take off? did notuse because it is super slow or doesn thave memory
    #     :return:
    #     """
    #     if not n_clusters:
    #         n_clusters = len(np.unique(self.Y_data))
    #
    #     self.clf = SpectralClustering(n_clusters=n_clusters, affinity=affinity,
    #                                   assign_labels=assign_labels,**params )
    #     self.clf.fit(self.X_data)
    #     self.y_labels = self.clf.labels_
    #     self.affinity_matrix = self.clf.affinity_matrix_
    #
    #     self.y_kmeans = self.clf.predict(self.X_data)
    #     print('accuracy', accuracy_score(self.Y_data, self.y_kmeans))
    #
    #     # table
    #     table = self.kmeans_table()
    #     print(table)
    #     sb = self.kmeans_seaborn_table()
    #
    # @timer
    # def gaussianmm(self,n_components = None,covariance_type='full', random_state=42, **params):
    #     """
    #     CHECK WHAT MAKES SENSE
    #     think not scalable for large datasets. so test in other
    #     https://scikit-learn.org/stable/modules/mixture.html#mixture
    #     https://jakevdp.github.io/PythonDataScienceHandbook/05.12-gaussian-mixtures.html
    #     https://scikit-learn.org/stable/modules/mixture.html
    #     https://scikit-learn.org/stable/auto_examples/mixture/plot_gmm_selection.html#sphx-glr-auto-examples-mixture-plot-gmm-selection-py
    #     :return:
    #     """
    #     if n_components is None:
    #         if self.Y_data:
    #             n_components=len(np.unique(self.Y_data))
    #         else:
    #             n_components = 1
    #     gmm = mixture.GMM(n_components=n_components, covariance_type=covariance_type, random_state=random_state, **params)
    #
    #     # the number of components can be ptimze by decreasing the BIC, see scikit learn and other page. graphics to do the est n_compoennts
    #     gmm_fit = gmm.fit(self.X_data)
    #     labels = gmm.predict(self.X_data)
    #     probs = gmm.predict_proba(self.X_data)
    #     print(probs[:5].round(3))
    #
    # @timer
    # def bayesiangaussianmm(self,n_components = None,covariance_type='full', random_state=42, **params):
    #     """
    #     CHECK WHAT MAKES SENSE
    #     ssame. not scalable I think
    #     https://jakevdp.github.io/PythonDataScienceHandbook/05.12-gaussian-mixtures.html
    #     https://scikit-learn.org/stable/modules/mixture.html
    #     https://scikit-learn.org/stable/auto_examples/mixture/plot_gmm_selection.html#sphx-glr-auto-examples-mixture-plot-gmm-selection-py
    #     https://scikit-learn.org/stable/modules/generated/sklearn.mixture.BayesianGaussianMixture.html#sklearn.mixture.BayesianGaussianMixture
    #     this one can have less than n_compoennts. it makes aditional operations. so n_components is not the real number of components but the maximum
    #     :return:
    #     """
    #     if n_components is None:
    #         if self.Y_data:
    #             n_components=len(np.unique(self.Y_data))
    #         else:
    #             n_components = 1
    #     gmm = mixture.BayesianGaussianMixture(n_components=n_components, covariance_type=covariance_type, random_state=random_state, **params)
    #     gmm_fit = gmm.fit(self.X_data)
    #     labels = gmm.predict(self.X_data)
    #     probs = gmm.predict_proba(self.X_data)
    #     print(probs[:5].round(3))
    # def dbscan(self):
    #     # The DBSCAN algorithm views clusters as areas of high density separated by areas of low density. Due to this
    #     # rather generic view, clusters found by DBSCAN can be any shape, as opposed to k-means which assumes that clusters
    #     # are convex shaped. The central component to the DBSCAN is the concept of core samples, which are samples that are
    #     # in areas of high density.
    #     clustering = OPTICS(min_samples=2,cluster_method='dbscan').fit(self.X_data)
    #     self.labels = clustering.labels_
    #     hierarchy = clustering.cluster_hierarchy_
    #     return clustering, self.labels

    # def kmeans_predict(self, x_test, output='add', max_iter=300, n_clusters=None):
    #     """
    #     Perform the kmeans to train data and predict the tests set
    #     :param output: 'add' or 'replace'.
    #     If add, the labels produced by clustering will be added as features
    #     If replace, labels produced will be replace the old labels
    #     :param max_iter: max number of iterations of cluster
    #     :param n_clusters: if None, it will define the number of clusters as the number of existing labels
    #     :return: values of X datasets altered (if add) or the Y datasets replaced
    #     TEST IT BECAUSE I MAKE CHANGES
    #     """
    #     if not n_clusters: n_clusters = len(np.unique(self.Y_data))
        # check if model built or no
        # if no
        #     clf = KMeans(n_clusters=n_clusters, max_iter=max_iter, random_state=42)
        #     clf.fit(self.X_data)
        #
        # if yes
        #     get the model
        #     predict the model to x_test
        #     y_labels_test = clf.predict(x_test)
        # get the labels of model (trained here or in class)
        # y_labels_train = clf.labels_
        # define what happens to outputs
        # if output == 'add':
        #     self.X_train.loc[:, 'km_clust'] = y_labels_train
        #     self.X_test.loc[:, 'km_clust'] = y_labels_test
        # elif output == 'replace':
        #     self.X_train = y_labels_train.loc[:, np.newaxis]
        #     self.X_test = y_labels_test.loc[:, np.newaxis]
        # else:
        #     raise ValueError('output should be add or replace')
        # return self

    # def classify(self, model=SVC(random_state=42)):
    #     """
    #     Function that fits the model in train datasets and predict on the tests dataset, returning the accuracy
    #     :param model: model to make prediction (SVC by default)
    #     :return: the accuracy of the prediction
    #     """
    #
    #     model.fit(self.X_train, self.y_train)
    #     y_pred = model.predict(self.X_test)
    #     print('Accuracy: {}'.format(accuracy_score(self.y_test, y_pred)))
    # take off dont know the sense of this here

# https://jakevdp.github.io/PythonDataScienceHandbook/05.11-k-means.html
# implement with tensorflow for GPU
# https://www.altoros.com/blog/using-k-means-clustering-in-tensorflow/
# https://www.commencis.com/thoughts/comparison-of-clustering-performance-for-both-cpu-and-gpu/ KMEANS TF



# -*- coding: utf-8 -*-
"""
##############################################################################

A class used for utility functions and transformer classes to change raw feature
vectors into a representation that is more suitable for the downstream estimators
All the functions are imported from sklearn.preprocessing

Authors: Ana Marta Sequeira

Date: 05/2019

Email:

##############################################################################
"""
from sklearn.preprocessing import *
from sklearn.feature_selection import VarianceThreshold
import pandas as pd
import numpy as np


class Preprocess:
    """
    The Preprocess class aims to transform the feature vectors into a representation suitable for downstream estimators.
    Clean the dataset from features redundants and deal with Nans.
    All the functions are imported from the module sklearn.preprocessing
    """

    def __init__(self):
        """	constructor """

    def missing_data(self, data):
        """
        check nans. if yes: returns a error message

        :param data: dataframe
        :return: error message or 0 nans error
        """

        if len(data.columns[data.isnull().any()]) != 0:
            print('warning, miss values. should drop columns or fill this values')
        else:
            print('0 nans')
        # print(dataset.columns[dataset.isnull().any()])
        # print(dataset.isnull().values.any())

    def remove_columns_all_zeros(self,data, columns_names=False):
        """
        Removes columns that have all values as zero.

        :param data: dataframe
        :param columns_names: if True retrieves the names of columns with only zeros
        :return: dataset without columns with all values=zero
        """
        if columns_names:
            return data.loc[:, (~(data == 0).all())], data.columns[(data == 0).all()]
        else:
            return data.loc[:, (~(data == 0).all())],None

    def remove_duplicate_columns(self, data, columns_names=False):
        """
        Removes columns duplicated.

        :param data: dataframe
        :param columns_names: if True retrieves the names of columns duplicates
        :return: dataset without duplicated columns
        """

        if columns_names:
            return data.loc[:, ~data.T.duplicated(keep='first')], data.columns[
                (data.T.duplicated(keep='first'))]
        else:
            return data.loc[:, ~data.T.duplicated(keep='first')], None
        # return dataset.T.drop_duplicates().T

    def remove_low_variance(self, data, threshold=0, standard=True, columns_names=False):
        """
        Based on scikit learn
        VarianceThreshold is a simple baseline approach to feature selection.
        It removes all features whose variance doesn’t meet some threshold

        :param data: dataframe
        :param threshold: value. The threshold of variance to drop columns (eg 0.8)
        :param standard:
            if ,in the case of threshold >0, the user wants to standardize features before apply variance
            threshold. minmaxscaler will be applied
        :param columns_names:
        :return: dataset without low variance columns (not scaled)
        """

        data = data
        if standard:
            scaler = MinMaxScaler()
            scaler.fit(data)
            scaler.transform(data)
            sel = VarianceThreshold(threshold)
            transf = sel.fit_transform(data)

            # original dataset without columns
            column_selected = sel.get_support(indices=True)

            data = data.iloc[:, column_selected]
            columns_excluded = list(set(data.columns) - set(data.columns))
            # print(data.head)
            # print(data.describe)

            if columns_names:
                return data, columns_excluded
            else:
                return data

        else:
            sel = VarianceThreshold(threshold)
            transf = sel.fit_transform(data)
            # original dataset without columns
            column_selected = sel.get_support(indices=True)
            data = data.iloc[:, column_selected]
            columns_excluded = list(set(data.columns) - set(data.columns))

            if columns_names:
                return transf, columns_excluded
            else:
                return transf

    def preprocess(self, data, columns_names=True, threshold=0, standard=True):
        """
        Removes columns that have all values as zero, duplicated and low variance columns

        :param data: dataset of input
        :param columns_names: if True retrieves the names of columns deleted
        :param threshold: the threshold of variance to drop columns
        :param standard: if true minmaxscaler will be applied
        :return: dataset original without deleted columns
        """

        dataset_zero, column_zero = self.remove_columns_all_zeros(data, columns_names=columns_names)
        dataset_without_duplicate, column_duplicated = self.remove_duplicate_columns(dataset_zero, columns_names=columns_names)
        dataset_clean, column_not_variable = self.remove_low_variance(dataset_without_duplicate, threshold=threshold,
                                                                      standard=standard, columns_names=columns_names)
        columns_deleted = (column_zero.append(column_duplicated)).append(column_not_variable)
        if columns_names:
            return dataset_clean, columns_deleted
        else:
            return dataset_clean




import numpy as np
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
# from silence_tensorflow import silence_tensorflow
# silence_tensorflow()
# import logging
# logging.disable(logging.WARNING)
# logging.getLogger("tensorflow").setLevel(logging.FATAL)
import tensorflow as tf

from tensorflow.keras.models import Sequential, load_model
from tensorflow.keras.layers import Dense, Dropout, LSTM, Embedding, TimeDistributed, Bidirectional
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.keras import regularizers, Input
from tensorflow.keras.layers import BatchNormalization, Flatten
from tensorflow.keras.callbacks import ReduceLROnPlateau, EarlyStopping, ModelCheckpoint
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import cross_val_score
from tensorflow.keras.preprocessing.sequence import pad_sequences

from tensorflow.keras.utils import plot_model
from tensorflow import keras
from operator import itemgetter

tf.debugging.set_log_device_placement(True)
strategy = tf.distribute.MirroredStrategy()


############################################################################################
############################################################################################


def create_dnn_simple(input_dim, number_classes,
                      hidden_layers=(128, 64, 32),
                      optimizer='Adam',
                      initial_dropout_value=0.0,
                      dropout_rate=(0.3,),
                      batchnormalization=(True,), activation="relu",
                      l1=1e-5, l2=1e-4, loss_fun='binary_crossentropy', activation_fun='sigmoid'):
    last_layer = hidden_layers[-1]
    # todo if give dropout rate not equal to layers, add?
    if len([dropout_rate]) == 1:
        dropout_rate = list(dropout_rate * len(hidden_layers))
    if len([batchnormalization]) == 1:
        batchnormalization = list(batchnormalization * len(hidden_layers))

    with strategy.scope():
        model = Sequential()
        model.add(Input(shape=(input_dim,)))

        if initial_dropout_value > 0:
            model.add(Dropout(initial_dropout_value))

        for layer in range(len(hidden_layers)):
            model.add(Dense(units=hidden_layers[layer], activation=activation,
                            kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            if batchnormalization[layer]:
                model.add(BatchNormalization())
            model.add(Dropout(dropout_rate[layer]))

        # # last layer
        # model.add(Dense(units=last_layer, activation=activation, kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
        # if batchnormalization: model.add(BatchNormalization())
        # if final_dropout_value > 0: model.add(Dropout(final_dropout_value))

        # Add Classification Dense, Compile model and make it ready for optimization
        model.add(Dense(number_classes, activation=activation_fun))
        model.compile(loss=loss_fun, optimizer=optimizer, metrics=['accuracy'])

        return model


def create_dnn_embedding(input_dim, number_classes,
                         hidden_layers=(128, 64),
                         optimizer='Adam',
                         input_dim_emb=21, output_dim=256, input_length=1000, mask_zero=True,
                         dropout_rate=(0.3,),
                         batchnormalization=True, activation="relu",
                         l1=1e-5, l2=1e-4,
                         loss_fun='binary_crossentropy', activation_fun='sigmoid'):
    if len([dropout_rate]) == 1:
        dropout_rate = list(dropout_rate * len(hidden_layers))
    if len([batchnormalization]) == 1:
        batchnormalization = list(batchnormalization * len(hidden_layers))

    with strategy.scope():
        model = Sequential()
        model.add(Input(shape=(input_dim,)))

        model.add(
            Embedding(input_dim=input_dim_emb, output_dim=output_dim, input_length=input_length, mask_zero=mask_zero))
        model.add(Flatten(data_format=None))

        for layer in range(len(hidden_layers)):
            model.add(Dense(units=hidden_layers[layer], activation=activation,
                            kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            if batchnormalization[layer]:
                model.add(BatchNormalization())
            model.add(Dropout(dropout_rate[layer]))

        # Add Classification Dense, Compile model and make it ready for optimization
        model.add(Dense(number_classes, activation=activation_fun))
        model.compile(loss=loss_fun, optimizer=optimizer, metrics=['accuracy'])

        return model



import pandas as pd
import numpy as np
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
# from silence_tensorflow import silence_tensorflow
# silence_tensorflow()
# import logging
# logging.disable(logging.WARNING)
# logging.getLogger("tensorflow").setLevel(logging.FATAL)

import tensorflow as tf
from tensorflow import keras

from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Embedding, Bidirectional
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.keras import regularizers
from tensorflow.keras.callbacks import ReduceLROnPlateau, EarlyStopping, ModelCheckpoint
from tensorflow.keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.layers import LSTM, ConvLSTM2D, BatchNormalization
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import MaxPool1D, MaxPool3D,TimeDistributed
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Input

tf.debugging.set_log_device_placement(True)
strategy = tf.distribute.MirroredStrategy()


def create_lstm_bilstm_simple(input_dim, number_classes,
                              optimizer='Adam',
                              bilstm=True,
                              lstm_layers=(128, 64),
                              activation='tanh',
                              recurrent_activation='sigmoid',
                              dropout_rate=(0.3,), recurrent_dropout_rate=(0.3,),
                              l1=1e-5, l2=1e-4,
                              dense_layers=(64,32),
                              dropout_rate_dense=(0.3,),
                              batchnormalization = (True,),
                              dense_activation="relu", loss_fun='binary_crossentropy', activation_fun='sigmoid'):

    if len([dropout_rate]) == 1:
        dropout_rate = list(dropout_rate * len(lstm_layers))
    if len([recurrent_dropout_rate]) == 1:
        recurrent_dropout_rate = list(recurrent_dropout_rate * len(lstm_layers))
    if len([dropout_rate_dense]) == 1:
        dropout_rate_dense = list(dropout_rate_dense * len(dense_layers))
    if len([batchnormalization]) == 1:
        batchnormalization = list(batchnormalization * len(dense_layers))


    last_lstm_layer = lstm_layers[-1]
    first_lstm_layer = lstm_layers[0]
    middle_lstm_layer = lstm_layers[1:-1]
    last_dense_layer = dense_layers[-1]
    middle_lstm_dropout=dropout_rate[1:-1]
    middle_lstm_recurrent_dropout=recurrent_dropout_rate[1:-1]


    with strategy.scope():
        model = Sequential()
        model.add(Input(shape=(1, input_dim,), dtype='float32', name='main_input'))

        # # add initial dropout
        # if int(initial_dropout_value) > 0:
        #     model.add(Dropout(initial_dropout_value))

        if not bilstm:
            # add first lstm layers
            model.add(LSTM(units=first_lstm_layer, input_shape=(1, input_dim,), return_sequences=True,
                           activation=activation, recurrent_activation=recurrent_activation,
                           dropout=dropout_rate[0],
                           recurrent_dropout=recurrent_dropout_rate[0],
                           kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))

            # add other lstm layers
            for layer in range(len(middle_lstm_layer) - 1):
                model.add(LSTM(units=middle_lstm_layer[layer], return_sequences=True,
                               activation=activation, recurrent_activation=recurrent_activation,
                               dropout=middle_lstm_dropout[layer],
                               recurrent_dropout=middle_lstm_recurrent_dropout[layer],
                               kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            # add last lstm layer
            model.add(LSTM(units=last_lstm_layer, return_sequences=False,
                           activation=activation, recurrent_activation=recurrent_activation,
                           dropout=dropout_rate[-1],
                           recurrent_dropout=recurrent_dropout_rate[-1],
                           kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))

        elif bilstm:
            # add first bilstm layer
            model.add(Bidirectional(LSTM(units=first_lstm_layer, return_sequences=True, activation=activation,
                                         recurrent_activation=recurrent_activation,
                                         kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2),
                                         dropout=dropout_rate[0], recurrent_dropout=recurrent_dropout_rate[0]),
                                    input_shape=(1, input_dim,)))

            # add other lstm layer
            for layer in range(len(middle_lstm_layer) - 1):
                model.add(Bidirectional(LSTM(units=middle_lstm_layer[layer], return_sequences=True,
                                             activation=activation, recurrent_activation=recurrent_activation,
                                             kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2),
                                             dropout=middle_lstm_dropout[layer], recurrent_dropout=middle_lstm_recurrent_dropout[layer])))
            # add last lstm layer
            model.add(Bidirectional(LSTM(units=last_lstm_layer, return_sequences=False,
                                         activation=activation, recurrent_activation=recurrent_activation,
                                         kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2),
                                         dropout=dropout_rate[-1], recurrent_dropout=recurrent_dropout_rate[-1])))

        # add denses
        for layer in range(len(dense_layers)):
            model.add(Dense(units=dense_layers[layer], activation=dense_activation,
                        kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            if batchnormalization[layer]:
                model.add(BatchNormalization())
            model.add(Dropout(dropout_rate_dense[layer]))

        # Add Classification Dense, Compile model and make it ready for optimization
        model.add(Dense(number_classes, activation=activation_fun))
        model.compile(loss=loss_fun, optimizer=optimizer, metrics=['accuracy'])
        model._layers = [layer for layer in model._layers if not isinstance(layer, dict)]
        # Without the line above AttributeError: 'dict' object has no attribute 'name'
        # https://github.com/tensorflow/tensorflow/issues/38988
        return model


def create_lstm_embedding(number_classes,
                          optimizer='Adam',
                          input_dim_emb=21, output_dim=128, input_length=1000, mask_zero=True,
                          bilstm=True,
                          lstm_layers=(128, 64),
                          activation='tanh',
                          recurrent_activation='sigmoid',
                          dropout_rate=(0.3,), recurrent_dropout_rate=(0.3,),
                          l1=1e-5, l2=1e-4,
                          dense_layers=(64, 32),
                          dense_activation="relu",
                          dropout_rate_dense = (0.3,),
                          batchnormalization = (True,),
                          loss_fun='binary_crossentropy', activation_fun='sigmoid'):

    if len([dropout_rate]) == 1:
        dropout_rate = list(dropout_rate * len(lstm_layers))
    print(dropout_rate)

    if len([recurrent_dropout_rate]) == 1:
        recurrent_dropout_rate = list(recurrent_dropout_rate * len(lstm_layers))
    if len([dropout_rate_dense]) == 1:
        dropout_rate_dense = list(dropout_rate_dense * len(dense_layers))
    print(dropout_rate_dense)
    if len([batchnormalization]) == 1:
        batchnormalization = list(batchnormalization * len(dense_layers))


    last_lstm_layer = lstm_layers[-1]

    with strategy.scope():
        model = Sequential()
        # model.add(Input(shape=(input_dim,),dtype='float32', name='main_input'))
        # Add an Embedding layer expecting input vocab of size 1000, and
        model.add(
            Embedding(input_dim=input_dim_emb, output_dim=output_dim, input_length=input_length, mask_zero=mask_zero))
        # model.add(Flatten(data_format=None))
        # model.add(tf.keras.layers.Reshape((input_length,output_dim)))
        # print(model.output_shape) #None 100, 256
        last_layer = LSTM(units=last_lstm_layer, return_sequences=False,
                          activation=activation, recurrent_activation=recurrent_activation,
                          dropout=dropout_rate[-1],
                          recurrent_dropout=recurrent_dropout_rate[-1],
                          kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2))
        if not bilstm:
            # add lstm layers
            for layer in range(len(lstm_layers) - 1):
                model.add(LSTM(units=lstm_layers[layer], return_sequences=True,
                               activation=activation, recurrent_activation=recurrent_activation,
                               dropout=dropout_rate[layer],
                               recurrent_dropout=recurrent_dropout_rate[layer],
                               kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            # add last lstm layer
            model.add(last_layer)

        elif bilstm:
            # add other lstm layer
            for layer in range(len(lstm_layers) - 1):
                model.add(Bidirectional(LSTM(units=lstm_layers[layer], return_sequences=True,
                                             activation=activation, recurrent_activation=recurrent_activation,
                                             kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2),
                                             dropout=dropout_rate[layer], recurrent_dropout=recurrent_dropout_rate[layer])))
            # add last lstm layer
            model.add(Bidirectional(last_layer))

        # add denses
        for layer in range(len(dense_layers)):
            model.add(Dense(units=dense_layers[layer], activation=dense_activation,
                            kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            if batchnormalization[layer]:
                model.add(BatchNormalization())
            model.add(Dropout(dropout_rate_dense[layer]))
        # Add Classification Dense, Compile model and make it ready for optimization
        # model binary
        model.add(Dense(number_classes, activation=activation_fun))
        model.compile(loss=loss_fun, optimizer=optimizer, metrics=['accuracy'])
        return model

#
# def create_conv_lstm(number_classes,
#                      maxlen,
#                      input_dim,
#                      optimizer='Adam',
#                      bilstm=True,
#                      filters=(128, 64),
#                      dense_layers=(64, 32),
#                      activation='tanh',
#                      recurrent_activation='sigmoid',
#                      dense_activation="relu",
#                      max_pooling=True,
#                      pool_size=2, strides_pool=1,
#                      data_format_pool='channels_first',
#                      padding='same',
#                      dropout_rate=0.3, recurrent_dropout_rate=0.3,
#                      l1=1e-5, l2=1e-4,
#                      final_dropout_value=0.0):
#     # https://www.kaggle.com/kcostya/convlstm-convolutional-lstm-network-tutorial
#     # https://medium.com/neuronio-br/uma-introdu%C3%A7%C3%A3o-a-convlstm-c14abf9ea84a
#     # https://www.nature.com/articles/s41598-019-46850-0
#     # todo so tem conv2D , mas ate faz sentido. ver se no cnn_lstm tb n fazia diferenca fazer c 2D. ou o stride. pq s o stride for 20 n tiramos nada de la, pq vai sempre buscar o numero 1 q é  o mais alto
#     # ver papaers q usem cnn stm ou convlstm c sequencias proteicas paratentar perceber melhor isto
#     # todo perceber melhor e melhorar a aquitectura do cnn lstm
#     # http://colah.github.io/posts/2015-08-Understanding-LSTMs/
#     # https://keras.io/api/layers/recurrent_layers/time_distributed/
#
#     # https://github.com/keras-team/keras/issues/6150
#     # https://keras.io/api/layers/recurrent_layers/conv_lstm2d/
#     # https://medium.com/neuronio/an-introduction-to-convlstm-55c9025563a7
#
#     last_lstm_layer = filters[-1]
#     last_dense_layer = dense_layers[-1]
#     # maxlen=x_train.shape[1]
#     # input_dim=x_train.shape[2]
#     # trailer_input  = Input(shape=(frames, channels, pixels_x, pixels_y)
#     with strategy.scope():
#         model = Sequential()
#         model.add(Input(shape=(1,input_dim, 1, 1)))
#         # If data_format='channels_first' 5D tensor with shape: (samples, time, channels, rows, cols)
#         # If data_format='channels_last' 5D tensor with shape: (samples, time, rows, cols, channels)
#         print(model.input_shape)
#         for layer in range(len(filters) - 1):
#             model.add(ConvLSTM2D(filters=filters[layer],
#                                  data_format='channels_last',
#                                  kernel_size=(1, 1),
#                                  strides=(1, 1),
#                                  padding='same',
#                                  dilation_rate=(1, 1),
#                                  activation='tanh',
#                                  recurrent_activation='hard_sigmoid',
#                                  use_bias=True,
#                                  kernel_initializer='glorot_uniform',
#                                  recurrent_initializer='orthogonal',
#                                  bias_initializer='zeros',
#                                  unit_forget_bias=True,
#                                  kernel_regularizer=None,
#                                  recurrent_regularizer=None,
#                                  bias_regularizer=None,
#                                  activity_regularizer=None,
#                                  kernel_constraint=None,
#                                  recurrent_constraint=None,
#                                  bias_constraint=None,
#                                  return_sequences=True,
#                                  go_backwards=False,
#                                  stateful=False,
#                                  dropout=0.0,
#                                  recurrent_dropout=0.0))
#             model.add(BatchNormalization())
#             model.add(MaxPool3D(pool_size=pool_size, strides=strides_pool, padding=padding,
#                                              data_format=data_format_pool))
#         model.add(TimeDistributed(Flatten()))
#
#         model.add(TimeDistributed(Dense(512,)))
#         model.add(TimeDistributed(Dense(32,)))
#         model.add(TimeDistributed(Dense(last_dense_layer)))
#         # model.add(Flatten())
#         # # add final dense
#         # for layer in range(len(dense_layers) - 1):
#         #     model.add(Dense(units=dense_layers[layer], activation=dense_activation,
#         #                     kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
#         #     model.add(Dropout(dropout_rate))
#         #
#         # # add last layer
#         # model.add(Dense(units=last_dense_layer, activation=dense_activation,
#         #                 kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
#         # # add final dropout
#         # if final_dropout_value > 0: model.add(Dropout(final_dropout_value))
#
#         # Add Classification Dense, Compile model and make it ready for optimization
#         # model binary
#         # if not number_classes > 2:
#         #     model.add(Dense(1, activation='sigmoid'))
#         #     model.compile(loss='binary_crossentropy', optimizer=optimizer, metrics=['accuracy'])
#         #
#         # # multiclass model
#         # elif number_classes > 2:
#         #     print('multiclass')
#         #     model.add(Dense(number_classes, activation='softmax'))
#         #     model.compile(loss='sparse_categorical_crossentropy', optimizer=optimizer, metrics=['accuracy'])
#         model.compile(loss='sparse_categorical_crossentropy', optimizer=optimizer, metrics=['accuracy'])
#         return model
#
# # model.add(Dense(y_train.shape[1], activation='sigmoid'))





# units=units_lstm, input_shape=(1, 2003),return_sequences=True, activation='tanh', recurrent_activation='sigmoid',
# use_bias=True, kernel_initializer='glorot_uniform', recurrent_initializer='orthogonal',
# bias_initializer='zeros', unit_forget_bias=True, kernel_regularizer=None, recurrent_regularizer=None,
# bias_regularizer=None, activity_regularizer=None, kernel_constraint=None, recurrent_constraint=None,
# bias_constraint=None, dropout=dropout_rate, recurrent_dropout=0.5, implementation=2, return_state=False,
# go_backwards=False, stateful=False, unroll=False


# -*- coding: utf-8 -*-
"""
#####################################################################################

Allows to change the sequence and generating other.cut or add aminoacids to obtain sequences with equal lenght.
(one sequence ----> one sequence).
	The function receives a sequence, the desired number of aa from the n terminal and from the c terminal.
	It receives a third argument that is called terminal. This argument is used, to decide in which direction add dummy aa
	(if needed)

	To considerer a protein only from the n terminal, c terminal=0 and the proteins will be cutted from left right.
	If necessary dummy will be added to the right

	To considerer a protein only from the c terminal, n terminal= 0 and the proteins will be cutted from right left.
	If necessary dummy will be added to the left

	To consider both extremes of sequence both n and c terminal are different of zero.
	If len of protein sequence is smaller than the both extremes together, the middle aa will be repeated.
	If necessary dummy aa will be added to the right (terminal=0), to the left (terminal=-1) or in the middle (terminal=2).


Authors:

Date:

Email:

#####################################################################################

"""

import re
import string

AALetter=["A","R","N","D","C","E","Q","G","H","I","L","K","M","F","P","S","T","W","Y","V"]
#############################################################################################


def seq_equal_lenght(seq,n_terminal=10, c_terminal=10, terminal=0, dummie='X'):
    """
    cut or add aminoacids to obtain sequences with equal lenght.
    :param seq: protein sequence
    :param n_terminal: number of aa to consider in the n terminal (left side of sequence)
    :param c_terminal: number of aa to consider in the c terminal (right side of sequence)
    :param terminal: in case of the need to add dummy aa and no terminal as already been chosen, it decides where to add
                    0 to add to the right (consider N terminal)
                    -1 to add to the left (consider C terminal)
                    2 to add in the middle (N and C terminal will be both present and repeated with dummy in middle
    :param dummie: dummie variable to add. X by default
    :return: list of sequences cntaining all the same lenght
    """

    size=len(seq)
    lenght=n_terminal+c_terminal
    if size>lenght:
        if c_terminal!=0:
            seq=seq[:n_terminal]+seq[-c_terminal:]

        else:
            seq=seq[:n_terminal]

    else: #the size of sequence is less than the lenght desired and dummy aa is needed

        #if one of the terminals is 0 the user already set in each direction to add

        if c_terminal==0:
            seq=seq[:lenght] + str(dummie*(lenght-size))

        elif n_terminal==0:
            seq= str(dummie*(lenght-size)) +seq[-lenght:]

        else: #if both terminals are filled, the variable terminal decide in which direction to cut/add

            if terminal==0: #add dummy to the right
                seq=seq[:lenght] + str(dummie*(lenght-size))


            elif terminal==-1: #add dummy to the left
                seq= str(dummie*(lenght-size)) +seq[-lenght:]


            else:#duplicate aa in the middle retaining the terminals. If necessary add dummy in the middle
                if n_terminal<size and c_terminal<size:
                    seq=seq[:n_terminal]+seq[-c_terminal:]

                elif n_terminal<size and c_terminal>size:
                    seq=seq[:n_terminal]+str(dummie*(c_terminal-size))+seq[-c_terminal:]

                elif n_terminal>size and c_terminal<size:
                    seq=seq[:n_terminal]+str(dummie*(n_terminal-size))+seq[-c_terminal:]

                else:
                    seq=seq[:n_terminal]+str(dummie*(n_terminal-size))+str(dummie*(c_terminal-size))+seq[-c_terminal:]

    return seq


if __name__ == "__main__":
    print((seq_equal_lenght('AAVFNDRAT', 5,0)))
    print((seq_equal_lenght('AAVFNDRAT', 0,5)))
    print((seq_equal_lenght('AAVFNDRAT', 5,5)))

    print((seq_equal_lenght('AAVFNDRAT', 5,5,-1)))
    print((seq_equal_lenght('AAVFNDRAT', 10,10,2)))

    print((seq_equal_lenght('ATN', 5,0)))
    print(seq_equal_lenght('ATN', 0,5))
    print(seq_equal_lenght('ATN', 5,5,2))
    print(seq_equal_lenght('ATN', 5,5,-1))

    print((seq_equal_lenght('AAVFNDRAT', 15,5,2)))








# -*- coding: utf-8 -*-
"""
##############################################################################

A class used for reading sequences or change sequences.
The main objective is to create sequence objects to calculate descriptors
The class allows to:
     1)Read sequences from string or from uniprot ID (is also possible retrieve sequences fom txt with uniprot IDs)
     2)Check if the protein sequence is a valid sequence
     3) Preprocessing methods to transform the protein sequences
     4)Obtain a sized sequence of list of sequences, adding or cutting from both n and c terminals
     5)From one sequence generate list of subsequences based on sliding window approach, from specific aa, from the
     the terminals or divide the sequence in parts

Authors:Ana Marta Sequeira, Miguel Barros

Date: 01/2019

Email:

##############################################################################
"""
from joblib import Parallel, delayed
from .adjuv_functions.sequence.get_sequence import get_protein_sequence, get_protein_sequence_from_txt
from .adjuv_functions.sequence.pro_check import protein_check
from .adjuv_functions.sequence.preprocess_seq import protein_preprocessing
from .adjuv_functions.sequence.get_sized_seq import seq_equal_lenght
from .adjuv_functions.sequence.get_sub_seq import sub_seq_sliding_window, sub_seq_to_aa, sub_seq_split,sub_seq_terminals
import pandas as pd


class ReadSequence:

    """
    The ReadSequence class aims to read the input and transform it into a sequence that can be used to calculate Descriptors.
    It accepts uniprotID, txt with uniprot ID and string with aminoacid sequence. The functions to read protein are based on the package pydpi.
    It allows to get sized sequences or a variety of subsequences from just one sequence.
    """

    def __init__(self):
        """	constructor """
        pass

# #########################################
    # read string or ID and get protein sequence

    def read_protein_sequence(self, protein_sequence=""):
        """
        Read a protein sequence.
        :param protein_sequence: String with sequence
        :return: String with aa sequence
        """

        self.protein_sequence=str.strip(protein_sequence)
        # index=protein_check(protein_sequence)

        # if index==0:
        #     print("Error......")
        #     print("Please input a correct protein.")
        # else:
        return self.protein_sequence

    def get_protein_sequence_from_txt(self, path, openfile, savefile):
        """
        Function to retrieve sequences from a txt with uniprot ID. Does not retrieve a sequence object
        :param path: directory path containing ID file
        :param openfile:ID file ('name.txt')
        :param savefile: saved file with obtained protein sequences ('name2.txt')
        :return:File containing string of sequences
        """
        self.txt=get_protein_sequence_from_txt(path,openfile,savefile)
        return self.txt

    def get_protein_sequence_from_id(self, uniprotid=""):
        """
        Downloading a protein sequence by uniprot id.
        :param uniprotid: String with UniprotID
        :return: String with aa sequence
        """
        self.protein=get_protein_sequence(str.strip(uniprotid))
        index=protein_check(self.protein)

        if index==0:
            print("Error......")
            print("Please input a correct protein.")
        else:
            self.protein_sequence=self.protein
        return self.protein_sequence

##########################################
    # check protein

    def checkprotein(self, protein_sequence=""):
        """
        Check whether the protein sequence is a valid amino acid sequence or not.
        Just check. Not assign to the object.
        :param protein_sequence: protein sequence
        :return: error or valid message.
        """

        index=protein_check(protein_sequence)

        if index==0:
            print("Error......")
            print("Please input a correct protein.")
        else:
            print('sequence valid')

    ##########################################
    # preprocessing sequence

    def get_preprocessing(self, ProteinSequence: str, B :str ='N', Z : str = 'Q', U :str = 'C', O: str = 'K', J : str = 'I', X :str = ''):
        '''
        Transforms the protein sequence by replacing aminoacids like Asparagine (B),  Glutamine(Z), Selenocysteine (U) and
         Pyrrolysine (O), by default they are replaced for the closest aminoacid residue if is present in the sequence, Asparagine (N), Glutamine (Q),
         Cysteinen(C) and Lysine (K), respectively. It also removes the ambiguous aminoacid (X) and alters the ambiguous
         aminoacid (J), by default it is replaced for Isoleucine (I). Use with caution.

        :param ProteinSequence: Protein sequence
        :param B: One-letter aminoacid code to replace Asparagine (B). Default amino acid is Asparagine (N)
        :param Z: One-letter aminoacid code to replace Glutamine(Z). Default amino acid is Glutamine (Q)
        :param U: One-letter aminoacid code to replace Selenocysteine (U). Default amino acid is Cysteinen(C)
        :param O: One-letter aminoacid code to replace Pyrrolysine (O). Default amino acid is Lysine (K)
        :param J: One-letter aminoacid code to replace ambiguous aminoacid (J). Default amino acid is Isoleucine (I)
        :param X: One-letter aminoacid code to replace ambiguous aminoacid (X), by default it is removed.
        :return: transformed protein sequence
        '''
        return protein_preprocessing(ProteinSequence,B, Z, U, O, J,X)

##########################################
    # parallelized preprocessing sequence

    def par_preprocessing(self, dataset, col: str, B :str ='N', Z : str = 'Q', U :str = 'C', O: str = 'K', J : str = 'I', X :str = '', n_jobs : int = 4):
        '''
        Transforms the protein sequence in the dataset by replacing aminoacids like Asparagine (B),  Glutamine(Z), Selenocysteine (U) and
         Pyrrolysine (O), by default they are replaced for the closest aminoacid residue if is present in the sequence, Asparagine (N), Glutamine (Q),
         Cysteinen(C) and Lysine (K), respectively. It also removes the ambiguous aminoacid (X) and alters the ambiguous
         aminoacid (J), by default it is replaced for Isoleucine (I). Use with caution.

        :param dataset: Pandas dataframe containing the sequences
        :param col: Column where the sequences are present
        :param B: One-letter aminoacid code to replace Asparagine (B). Default amino acid is Asparagine (N)
        :param Z: One-letter aminoacid code to replace Glutamine(Z). Default amino acid is Glutamine (Q)
        :param U: One-letter aminoacid code to replace Selenocysteine (U). Default amino acid is Cysteinen(C)
        :param O: One-letter aminoacid code to replace Pyrrolysine (O). Default amino acid is Lysine (K)
        :param J: One-letter aminoacid code to replace ambiguous aminoacid (J). Default amino acid is Isoleucine (I)
        :param X: One-letter aminoacid code to replace ambiguous aminoacid (X), by default it is removed.
        :param n_jobs: number of CPU cores to use
        :return: transformed Pandas dataframe
        '''

        if isinstance(dataset, pd.DataFrame):
            res = Parallel(n_jobs=n_jobs)(delayed(protein_preprocessing)(seq, B, Z, U, O, J,X) for seq in dataset[col])
            dataset[col] = res
            return dataset
        else: raise Exception('Parameter dataframe must be a pandas dataframe')


##########################################
    # Get equal size sequences

    def get_sized_seq(self, sequences=[], n_terminal=10, c_terminal=10, terminal=0, dummie = 'X'):
        """
        cut or add aminoacids to obtain sequences with equal lenght.
        :param sequences: list containing protein sequences (string) or just a protein sequence
        :param n_terminal: number of aa to consider in the n terminal (left side of sequence)
        :param c_terminal: number of aa to consider in the c terminal (right side of sequence)
        :param terminal: in case of the need to add dummy aa and no terminal as already been chosen, it decides where to add
                        0 to add to the right (consider N terminal)
                        -1 to add to the left (consider C terminal)
                        2 to add in the middle (N and C terminal will be both present and repeated with dummy in middle
        :return: list of sequences containing all the same lenght. if just one sequence given it will return a string
        """
        if len(sequences)==1:
            seq=seq_equal_lenght(sequences[0],n_terminal,c_terminal,terminal, dummie=dummie)
            return seq
        else:
            equal_size_sequences=[] #list to store sequences with equal lenght
            for seq in sequences:
                seq_2=seq_equal_lenght(seq,n_terminal,c_terminal,terminal, dummie=dummie)
                equal_size_sequences.append(seq_2)
            return equal_size_sequences

##########################################
    # Generate subsequences

    def get_sub_seq_sliding_window(self, seq, window_size=20, gap=1, index=True):
        """
        sliding window of the protein given. It will generate a list of n sequences with
        lenght equal to the value of window and spaced a gap value. It can or not retrieve
        the indices of location of the subsequence in the original sequence.

        :param seq: protein sequence
        :param window_size: number of aminoacids to considerer, lenght of the subsequence. for default 20
        :param gap: gap size of the search of windows in sequence. default 1
        :param index: if true, return the indices of location of the subsequence in the original sequence
        :return: list with subsequences generated with or without a list of tuples with location of subsequences
         in original sequence
        """

        list_sliding_window=sub_seq_sliding_window(seq,window_size,gap,index)
        return list_sliding_window

    def get_sub_seq_to_aa(self, seq, ToAA='S', window=5):
        """
        Get all 2*window+1 sub-sequences whose center is ToAA in a protein

        :param seq:  protein sequence
        :param ToAA: central (query point) amino acid in the sub-sequence
        :param window:  span (number of amnoacids to go front and back from the ToAA
        :return: list form containing sub-sequences
        """

        list_subseq_to_aa=sub_seq_to_aa(seq, ToAA, window)
        return list_subseq_to_aa

    def get_sub_seq_split(self, seq, number_of_subseq):
        """
        Split the original seq in n number of subsequences.
        :param seq: protein sequence
        :param number_of_subseq: number of subsequences to divide the original seq
        :return: list with number_of_sequences sequences.

        """
        list_sub_seq_split=sub_seq_split(seq,number_of_subseq)
        return list_sub_seq_split

    def get_sub_seq_terminals(self, seq, N_terminal=5, C_terminal=5, rest=True):
        """
        Divide the sequence in the N terminal and  C terminal with sizes defined by the user. It returns
        a list with N and C terminal and rest of the sequence.
        By default the N terminal is considerer to be the beggining of the sequence (left)

        :param seq: protein sequence
        :param N_terminal: size of the N terminal to consider. If zero will not return
        :param C_terminal: size of the C terminal to consider. If zero will not return
        :param rest: If true will return the restant subsequence
        :return: list with N, C and rest of the sequence
        """
        list_sub_seq_term= sub_seq_terminals(seq, N_terminal, C_terminal,rest)
        return list_sub_seq_term




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
##############################################################################

File with util functions used in classes shallow Learning and DeepLearning
Authors: Ana Marta Sequeira

Date:12/2020

Email:

##############################################################################
"""
import functools
import time
import pickle
import numpy as np
import pandas as pd
from tensorflow.keras.models import Sequential, load_model
from sklearn.metrics import classification_report, confusion_matrix, accuracy_score,matthews_corrcoef,roc_auc_score,f1_score


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value
    return wrapper_timer




def validation(model,x_test,y_test):
    print('validation for test data')
    pred = model.predict(x_test)
    pred = np.reshape(pred, (pred.shape[0], 1)) # to put the shape as (number_samples,1)
    print("=== Confusion Matrix ===")
    print(confusion_matrix(y_test, pred))
    print("=== Classification Report ===")
    print(classification_report(y_test, pred))
    print('\n')
    print('test accuracy', accuracy_score(y_test, pred))
    print('f1 score', f1_score(y_test, pred))
    print('roc_auc', roc_auc_score(y_test, pred))
    print('mcc', matthews_corrcoef(y_test, pred))
    print('\n')

def validation_multiclass(model,x_test,y_test):
    print('validation for test data')
    print(x_test)
    pred = model.predict(x_test)
    pred = np.reshape(pred, (pred.shape[0], 1)) # to put the shape as (number_samples,1)
    print("=== Confusion Matrix ===")
    print(confusion_matrix(y_test, pred))
    print("=== Classification Report ===")
    print(classification_report(y_test, pred))
    print('\n')
    print('test accuracy', accuracy_score(y_test, pred))
    print('f1 score weighted', f1_score(y_test, pred, average='weighted'))
    print('f1 score macro', f1_score(y_test, pred, average='macro'))
    print('f1 score micro', f1_score(y_test, pred, average='micro'))
    # print('roc_auc ovr', roc_auc_score(y_test, pred, average= 'weighted',multi_class='ovr'))
    # print('roc_auc ovo', roc_auc_score(y_test, pred, average= 'weighted', multi_class='ovo'))
    print('mcc', matthews_corrcoef(y_test, pred))
    print('\n')


def saveModel(modelType, model, name):
    if modelType != 'dnn':
        pickle.dump(model, open(name, 'wb'))
    else:
        model.model.save(name + '.h5')


def loadDNN(modelname):
    model = load_model(modelname)
    model.compile(loss='binary_crossentropy',
                  metrics=['accuracy'])
    return model




def divide_dataset(file_in, columns_not_x=[], column_label=''):
    # separate labels
    # ec_number = dataset[EC_number]
    # uniprot = dataset[uniprot_ID]
    # sequence = dataset[sequence]
    dataset = pd.read_csv(file_in,low_memory=False)
    dataset_x = dataset.loc[:, dataset.columns[~dataset.columns.isin(columns_not_x)]]
    dataset_y=dataset.loc[:,column_label]
    return dataset, dataset_x, dataset_y
    #
    # x_original=dataset.loc[:, dataset.columns != 'labels']
    # labels=dataset['labels']


def binarize_labels(column_label):
    # print('binarize labels')
    fps_y_bin=[]
    for x in column_label:
        if x == 'non_Enzyme': fps_y_bin.append(int(0))
        elif x == 0 or x == '0': fps_y_bin.append(int(0))
        else: fps_y_bin.append(int(1))
    return fps_y_bin

def show_shapes(x_train, x_test, x_dval, y_train, y_test,y_dval):
    print("Expected: (num_samples, timesteps, channels)")
    print("x_train: {}".format(x_train.shape))
    print("x_test: {}".format(x_test.shape))
    print("x_dval: {}".format(x_dval.shape))

    print("y_train:   {}".format(y_train.shape))
    print("y_test:   {}".format(y_test.shape))
    print("y_dval:   {}".format(y_dval.shape))






#
#
# # def create_dnn(optimizer = 'Adam', dropout_rate=0.0,
# #              batchNormalization = True, embedding=True, hidden_layers=1,
# #              l1 = 0, l2 = 0, units = 512,units1 = 512, units2 = 512,
# #              nb_epoch=500,batch_size=10, input_dim=21, output_dim=100, input_length=None,mask_zero=True):
# def create_dnn(optimizer = 'Adam' , dropout_rate =0.0,
#                batchNormalization = True , embedding=True,hidden_layers =1,
#                l1 = 0, l2 = 0, units = 512 ,units1 = 512 , units2 = 512 ,
#                nb_epoch = 500, batch_size = 10, input_dim = 21, output_dim = 100, input_length = None,mask_zero = True):
#     print('dropout_rate = ', dropout_rate)
#
#     model = Sequential()
#     if embedding:
#         model.add(Embedding(input_dim=input_dim, output_dim=output_dim,input_length=input_length,mask_zero=mask_zero)) #todo input dim? output dim?
#         model.add(Flatten(data_format=None))
#
#     model.add(Dense(units=units, activation="relu"))
#     if batchNormalization:
#         model.add(BatchNormalization())
#
#     model.add(Dropout(dropout_rate))
#     for i in range(hidden_layers):
#         model.add(Dense(units=units, activation="relu", kernel_regularizer = regularizers.l1_l2(l1=l1, l2=l2)))
#         if batchNormalization:
#             model.add(BatchNormalization())
#         model.add(Dropout(dropout_rate))
#
#     model.add(Dense(1, activation='sigmoid'))
#
#     # #Compile model and make it ready for optimization
#     model.compile(loss='binary_crossentropy', optimizer = optimizer, metrics=['accuracy'])
#     # Reduce lr callback
#     # reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.5,patience=50, min_lr=0.00001, verbose=1)
#     return model
#
#
#
# @timer
# def train_dnn(X_train,y_train,X_test,y_test,x_dval,y_dval,
#              optimizer = 'Adam', dropout_rate=0.0,
#              batchNormalization = True, embedding=True, hidden_layers=1,
#              l1 = 0, l2 = 0, units = 512,units1 = 512, units2 = 512,
#              nb_epoch=500,batch_size=10, input_dim=1021, output_dim=100, input_length=400,mask_zero=True,
#              save = True, path = 'savedModels/modelDNN.h1', report = True):
#
#     print(input_dim, output_dim, input_length)
#     model=create_dnn(optimizer=optimizer, dropout_rate=dropout_rate, batchNormalization=batchNormalization,
#                      embedding=embedding,hidden_layers=hidden_layers, l1=l1, l2=l2, units = units, units1=units1,
#                      units2=units2,input_dim=input_dim, output_dim=output_dim, input_length=input_length,
#                      mask_zero=mask_zero, nb_epoch=nb_epoch, batch_size=batch_size)
#
#     print('=== DNN ===')
#     model = KerasClassifier(build_fn=create_dnn, verbose=0)
#     # model = KerasClassifier(build_fn=model,  dropout_rate = dropout_rate, hidden_layers = hidden_layers,
#     #                     l1 = l1, l2 = l2, units = units, verbose=0)
#     # simple early stopping
#     es = EarlyStopping(monitor='val_loss', mode='min', verbose=1,patience=10)
#
#     # reduce learning rate
#     reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.5,patience=50, min_lr=0.00001, verbose=1)
#
#     #checkpoint
#     filepath='weights.{epoch:02d}-{val_loss:.2f}.hdf5'
#     cp=ModelCheckpoint(filepath=filepath, monitor='val_loss', verbose=0, save_best_only=True, save_weights_only=False,
#                        mode='auto', period=1)
#
#     callbacks = [es, reduce_lr, cp]
#     # callbacks = [reduce_lr]
#
#     #Training
#     dnn = model.fit(X_train, y_train, epochs=nb_epoch, batch_size=batch_size, callbacks=callbacks,
#                     validation_data = (X_test, y_test),verbose=1)
#
#     if report:
#         print('Training Accuracy mean: ', np.mean(dnn.history['accuracy']))
#         print('Validation Accuracy mean: ', np.mean(dnn.history['val_accuracy']))
#
#         print('Training Loss mean: ', np.mean(dnn.history['loss']))
#         print('Validation Loss mean: ', np.mean(dnn.history['val_loss']))
#
#         summary_accuracy(dnn)
#         summary_loss(dnn)
#
#
#     #todo por aqui os pesos do melhor modelo os parametros
#     # todo por a guardar so o melhor modelo e n de epoc em epoch
#     #todo confirmar q acho q esta a fazer mais hidden layers do q o suposto ou melhor,
#     # adiciona mais uma pelo hidden layers. n sei s a primeira conta como hidden ou nao
#
#     dnn_predict = model.predict(x_dval)
#     print("=== Confusion Matrix ===")
#     print(confusion_matrix(y_dval, dnn_predict))
#     print("=== Classification Report ===")
#     print(classification_report(y_dval, dnn_predict))
#     print('\n')
#     print("=== Test accuracy ===")
#     print(accuracy_score(y_dval, dnn_predict))
#     print('\n')
#
#     print(model.model.summary())
#
#     if save:
#         try:
#             model.model.save(path)
#         except:
#             print('did not save')
#             return model
#
#     return model
#
#
# ############################################ Recurrent neural networks ################################################
# #
# # keras.layers.LSTM(units, activation='tanh', recurrent_activation='sigmoid', use_bias=True,
# #                   kernel_initializer='glorot_uniform', recurrent_initializer='orthogonal', bias_initializer='zeros',
# #                   unit_forget_bias=True, kernel_regularizer=None, recurrent_regularizer=None, bias_regularizer=None,
# #                   activity_regularizer=None, kernel_constraint=None, recurrent_constraint=None, bias_constraint=None,
# #                   dropout=0.0, recurrent_dropout=0.0, implementation=2, return_sequences=False, return_state=False,
# #                   go_backwards=False, stateful=False, unroll=False)
# # https://keras.io/layers/recurrent/
#
# # https://adventuresinmachinelearning.com/keras-lstm-tutorial/
# #
#
#
# def create_lstm (input_dim, output_dim, input_length, batchNormalization, embedding, dropout_rate,
#                  optimizer,time_distibuted, hidden_layers, l1, l2, units_lstm, units1,
#                  units2, mask_zero,bidirectional=False):
#     model = Sequential()
#     # Add an Embedding layer expecting input vocab of size 1000, and
#     # output embedding dimension of size 64.
#     if embedding:
#         model.add(Embedding(input_dim=input_dim, output_dim=output_dim,input_length=input_length,mask_zero=mask_zero))
#         model.add(Flatten(data_format=None))
#
#     # Add a LSTM layer with 128 internal units.
#     if bidirectional:
#         model.add(Bidirectional(LSTM(units=units_lstm, return_sequences=True, dropout=0.5, recurrent_dropout=0.5)))
#         model.add(Bidirectional(LSTM(units=units_lstm, return_sequences=True, dropout=0.5, recurrent_dropout=0.5)))
#
#     else:
#         model.add(LSTM(units=units_lstm,return_sequences=True, activation='tanh', recurrent_activation='sigmoid',
#                        use_bias=True, kernel_initializer='glorot_uniform', recurrent_initializer='orthogonal',
#                        bias_initializer='zeros', unit_forget_bias=True, kernel_regularizer=None, recurrent_regularizer=None,
#                        bias_regularizer=None, activity_regularizer=None, kernel_constraint=None, recurrent_constraint=None,
#                        bias_constraint=None, dropout=0.5, recurrent_dropout=0.5, implementation=2, return_state=False,
#                        go_backwards=False, stateful=False, unroll=False))
#
#         model.add(LSTM(units=units_lstm, return_sequences=True, dropout=0.5, recurrent_dropout=0.5))
#
#     if batchNormalization:
#         model.add(BatchNormalization())
#
#     if time_distibuted:
#         model.add(TimeDistributed(Dense(input_dim, activation='softmax', name='main_output')))
#         # considering a batch of 32 samples,  where each sample is a sequence  of 10 vectors of 16 dimensions. The batch
#         # input shape of the layer is then(32, 10, 16), and the input_shape, not including the samples dimension, is (10, 16).
#         # TimeDistributed to samples apply a Dense layer to each of the 10 timesteps, independently
#
#     else:
#         model.add(Dense(1, activation='sigmoid'))
#
#     # #Compile model and make it ready for optimization
#     model.compile(loss='binary_crossentropy', optimizer = optimizer, metrics=['accuracy'])
#     # Reduce lr callback
#     # reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.5,patience=50, min_lr=0.00001, verbose=1)
#     model.summary()
#     return model
#
#     # metrics=['categorical_accuracy']
#
# @timer
# def train_lstm( X_train, y_train, X_test, y_test, x_dval, y_dval,
#                 optimizer='Adam', dropout_rate=0.0,
#                 batchNormalization=True, embedding=True, hidden_layers=1,
#                 l1=0, l2=0, units_lstm=512, units1=512, units2=512,
#                 nb_epoch=500, batch_size=10, input_dim=21, output_dim=100, input_length=None, mask_zero=True,
#                 time_distributed=False,
#                 save=True, path='savedModels/modelLSTM.h1', report=True):
#
#     #validation_data data on which to evaluate the loss and any model metrics at the end of each epoch.
#     # tuple (x_val, y_val) of Numpy arrays or tensors - tuple (x_val, y_val, val_sample_weights) of Numpy arrays - dataset or a dataset iterator
#
#     model= create_lstm (optimizer=optimizer, dropout_rate=dropout_rate, batchNormalization=batchNormalization,
#                         embedding=embedding, input_dim=input_dim, output_dim=output_dim, input_length=input_length,
#                         hidden_layers=hidden_layers, l1=l1, l2=l2, units_lstm=units_lstm, units1=units1,
#                         units2=units2, mask_zero=mask_zero, time_distibuted=time_distributed)
#
#     # input_dim=1000, output_dim=64, batchNormalization=True, dropout_rate=0.0,time_distibuted=False
#
#     print('=== LSTM ===') # todo mantive igual ao dnn, n sei se é diferente. se for igual juntar os train
#     model = KerasClassifier(build_fn=create_lstm, dropout_rate=dropout_rate, hidden_layers=hidden_layers,
#                             l1=l1, l2=l2, units=units_lstm, verbose=0)
#
#     # simple early stopping
#     es = EarlyStopping(monitor='val_loss', mode='min', verbose=1, patience=10)
#
#     # earlyStopping = EarlyStopping(monitor='val_weighted_accuracy', patience=5, verbose=1, mode='auto') deepaclstm
#
#     # reduce learning rate
#     reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.5, patience=50, min_lr=0.00001, verbose=1)
#
#     # checkpoint
#     filepath = 'weights.{epoch:02d}-{val_loss:.2f}.hdf5'
#     cp = ModelCheckpoint(filepath=filepath, monitor='val_loss', verbose=0, save_best_only=True,
#                          save_weights_only=False,
#                          mode='auto', period=1)
#
#     callbacks = [es, reduce_lr, cp]
#     # callbacks = [reduce_lr]
#
#     # Training
#     lstm = model.fit(X_train, y_train,epochs=nb_epoch, batch_size=batch_size, callbacks=callbacks,
#                     validation_data=(X_test, y_test), verbose=1)
#
#     if report:
#
#         print('Training Accuracy mean: ', np.mean(lstm.history['accuracy']))
#         print('Validation Accuracy mean: ', np.mean(lstm.history['val_accuracy']))
#
#         print('Training Loss mean: ', np.mean(lstm.history['loss']))
#         print('Validation Loss mean: ', np.mean(lstm.history['val_loss']))
#
#         summary_accuracy(lstm)
#         summary_loss(lstm)
#
#
#     lstm_predict = model.predict(x_dval)
#
#     print("=== Confusion Matrix ===")
#     print(confusion_matrix(y_dval, lstm_predict))
#     print("=== Classification Report ===")
#     print(classification_report(y_dval, lstm_predict))
#     print('\n')
#     print("=== Test accuracy ===")
#     print(accuracy_score(y_dval, lstm_predict))
#     print('\n')
#
#     print(model.model.summary())
#
#     if save:
#         try:
#             model.model.save(path)
#         except:
#             print('did not save')
#             return model
#
#     return model


#generators to feed model wit h data https://medium.com/@anuj_shah/creating-custom-data-generator-for-training-deep-learning-models-part-1-5c62b20cff26

# import numpy as np
# import pandas as pd
# import random
# import tensorflow as tf
# import time as tm
#
# INPUT_SHAPE=[3, 5]
# NUM_POINTS=20
# BATCH_SIZE=7
# EPOCHS=4
#
# def data_gen(num, in_shape):
#     for i in range(num):
#         x = np.random.rand(in_shape[0], in_shape[1])
#         y = random.randint(0,2)
#         yield x, y
#
# train = tf.data.transporter_substrate.from_generator(
#     generator=data_gen,
#     output_types=(tf.float32, tf.int32),
#     output_shapes=([None, INPUT_SHAPE[1]],()),
#     args=([NUM_POINTS, INPUT_SHAPE])
# )
#
# def create_model(input_shape):
#     model = tf.keras.models.Sequential([
#         tf.keras.layers.Dense(100, activation="tanh", input_shape=input_shape),
#         tf.keras.layers.LSTM(1, activation="tanh"),
#         tf.keras.layers.Dense(3, activation="softmax")
#     ])
#     return model
#
# model = create_model(input_shape=INPUT_SHAPE)
#
# model.compile(
#     optimizer=tf.keras.optimizers.Adam(learning_rate=1e-4, clipvalue=1.0),
#     loss= tf.keras.losses.SparseCategoricalCrossentropy()
# )
# print(model.summary())
# model.fit(train.batch(BATCH_SIZE), epochs=EPOCHS, verbose=2)
# model.evaluate(train, steps=None, verbose=1)


#LSTM
# http://colah.github.io/posts/2015-08-Understanding-LSTMs/
# https://towardsdatascience.com/the-fall-of-rnn-lstm-2d1594c74ce0
# https://towardsdatascience.com/bert-explained-state-of-the-art-language-model-for-nlp-f8b21a9b6270
# https://towardsdatascience.com/bert-for-dummies-step-by-step-tutorial-fb90890ffe03
# https://medium.com/inside-machine-learning/what-is-a-transformer-d07dd1fbec04
# https://towardsdatascience.com/transformers-141e32e69591
# https://distill.pub/2016/augmented-rnns/
# https://skymind.ai/wiki/attention-mechanism-memory-network
# https://bair.berkeley.edu/blog/2019/11/04/proteins/
# https://adventuresinmachinelearning.com/keras-lstm-tutorial/
#  https://www.tensorflow.org/guide/keras/rnn





import torch
from numpy import argmax
from numpy import vstack
from sklearn.metrics import accuracy_score, f1_score, matthews_corrcoef, confusion_matrix, recall_score, roc_auc_score


def test(device, model, test_loader):
    """
    Test the model.
    :param model: Model to be tested.
    :param device: Device to be used for testing.
    :param test_loader: Data loader for testing.
    :return: The accuracy, mcc and confusion matrix of the model.
    """
    model.eval()
    predictions, actuals = list(), list()
    with torch.no_grad():
        for (inputs, targets) in test_loader:
            inputs, targets = inputs.to(device), targets.to(device)
            yhat = model(inputs)
            yhat = yhat.cpu().detach().numpy()
            actual = targets.cpu().numpy()
            actual = actual.reshape((len(actual), 1))
            yhat = argmax(yhat, axis=1)
            yhat = yhat.reshape((len(yhat), 1))
            predictions.append(yhat)
            actuals.append(actual)

    predictions, actuals = vstack(predictions), vstack(actuals)
    acc = accuracy_score(actuals, predictions)
    mcc = matthews_corrcoef(actuals, predictions)
    roc_auc = roc_auc_score(actuals, predictions)
    f1 = f1_score(actuals, predictions)
    recall = recall_score(actuals, predictions)
    report = confusion_matrix(actuals, predictions)

    metrics = {
        'accuracy': acc,
        'mcc': mcc,
        'roc_auc': roc_auc,
        'f1': f1,
        'recall': recall,
        'confusion_matrix': report
    }
    return metrics



import os
import pickle
import random
import sys
import math
import numpy as np
from itertools import product
import torch
import json
from torch import nn
from ray import tune

ALPHABET = 'ACGT'
ALPHABET_CUT = 'ACGTN'
pairs = {
    'A': 'T',
    'T': 'A',
    'G': 'C',
    'C': 'G'
}

combinations = {
    'mlp': ['descriptor'],
    'mlp_half': ['descriptor'],
    'cnn': ['one_hot', 'chemical', 'kmer_one_hot'],
    'cnn_half': ['one_hot', 'chemical', 'kmer_one_hot'],
    'lstm': ['one_hot', 'chemical', 'kmer_one_hot'],
    'bi_lstm': ['one_hot', 'chemical', 'kmer_one_hot'],
    'gru': ['one_hot', 'chemical', 'kmer_one_hot'],
    'bi_gru': ['one_hot', 'chemical', 'kmer_one_hot'],
    'cnn_lstm': ['one_hot', 'chemical', 'kmer_one_hot'],
    'cnn_bi_lstm': ['one_hot', 'chemical', 'kmer_one_hot'],
    'cnn_gru': ['one_hot', 'chemical', 'kmer_one_hot'],
    'cnn_bi_gru': ['one_hot', 'chemical', 'kmer_one_hot']
}

def print_metrics(model_label, mode, data_dir, kmer_one_hot, class_weights, metrics):
    print("-" * 40)
    print("Results in test set: ")
    print("-" * 40)
    print("model:        ", model_label)
    print("mode:         ", mode)
    print("dataset:      ", data_dir.split("/")[-1])
    if mode == "kmer_one_hot":
        print("kmer_one_hot: ", kmer_one_hot)
    if "essential_genes" in data_dir:
        print("class_weights:", class_weights)
    print("-" * 40)

    for key in metrics:
        if key == 'confusion_matrix':
            print(f"{key:<{20}}= {metrics[key][0]}")
            print(f"{'':<{20}}  {metrics[key][1]}")
        else:
            print(f"{key:<{20}}= {metrics[key]:.3f}")

    print("-" * 40)

# -----------------------------------------------------------------------------
def seed_everything(seed=42):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    os.environ["CUBLAS_WORKSPACE_CONFIG"]=":4096:2"
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.enabled = False

def checker(sequence):
    """
    Checks if the input sequence is a valid DNA sequence.
    """
    return all(i in ALPHABET for i in sequence)

def checker_cut(sequence):
    """
    Checks if the input sequence is a valid DNA sequence. Includes the 'N' character as valid because it is used to fill the sequence to the right length.
    """
    return all(i in ALPHABET_CUT for i in sequence)

def normal_round(n):
    """
    Equivalent to python's round but its rounds up if it ends up with 0.5.
    """
    if n - math.floor(n) < 0.5:
        return math.floor(n)
    return math.ceil(n)


def normalize_dict(dic):
    """Normalize the value of a dictionary."""
    N = sum(dic.values())
    for key in dic:
        dic[key] = round(dic[key] / N, 3)
    return dic


def make_kmer_list(k):
    try:
        return ["".join(e) for e in product(ALPHABET, repeat=k)]
    except TypeError:
        print("TypeError: k must be an inter and larger than 0, alphabet must be a string.")
        raise TypeError
    except ValueError:
        print("TypeError: k must be an inter and larger than 0")
        raise ValueError


def make_kmer_dict(k):
    try:
        return {''.join(i): 0 for i in product(ALPHABET, repeat=k)}
    except TypeError:
        print("TypeError: k must be an inter and larger than 0, alphabet must be a string.")
        raise TypeError
    except ValueError:
        print("TypeError: k must be an inter and larger than 0")
        raise ValueError

def calculate_kmer_onehot(k):
    nucleotides = [''.join(i) for i in product(ALPHABET_CUT, repeat=k)]
    encoded = []
    for i in range(5 ** k):
        encoded.append(np.zeros(5 ** k).tolist())
        encoded[i][i] = 1.0
        
    return {nucleotides[i]: encoded[i] for i in range(len(nucleotides))}

def calculate_kmer_list(sequence, k):
    l = []
    for i in range(len(sequence) - k + 1):
        l.append(sequence[i:i+k])
    return l

def read_config(device, filename='config.json'):
    """
    Reads the configuration file and validates the values. Returns the configuration.
    """
    with open(filename) as f:
        config = json.load(f)
    
    # ------------------------------------ check if data_dir exists ------------------------------------
    current_path = os.getcwd()
    current_path = current_path.replace("/notebooks", "") # when running from notebook
    
    config['combination']['data_dir'] = current_path + config['combination']['data_dir']
    if not os.path.exists(config['combination']['data_dir']):
        raise ValueError("Data directory does not exist:", config['combination']['data_dir'])    

    # --------------------------- check if model and mode combination is valid -------------------------
    model_label = config['combination']['model_label']
    mode = config['combination']['mode']
    if(model_label in combinations):
        if(mode not in combinations[model_label]):
            raise ValueError(model_label, 'does not support', mode, ', please choose one of', combinations[model_label])
    else:
        raise ValueError('Model label:', model_label, 'not implemented in', combinations.keys())

    # --------------------------- check if it's binary classification ----------------------------------
    loss = config['fixed_vals']['loss_function']
    output_size = config['fixed_vals']['output_size']
    if(loss != "cross_entropy" or output_size != 2):
        raise ValueError(
            'Model is not binary classification, please set loss_function to cross_entropy and output_size to 2')

    # --------------------------- create the cross entropy pytorch object ------------------------------
    class_weights = torch.tensor(config['combination']['class_weights']).to(device)
    config['fixed_vals']['loss_function'] = nn.CrossEntropyLoss(weight=class_weights)
    
    # --------------------------- create ray tune objects ----------------------------------------------
    config['hyperparameter_search_space']["hidden_size"] = tune.choice(config['hyperparameter_search_space']['hidden_size'])
    config['hyperparameter_search_space']["lr"] = tune.choice(config['hyperparameter_search_space']['lr'])
    config['hyperparameter_search_space']["batch_size"] = tune.choice(config['hyperparameter_search_space']['batch_size'])
    config['hyperparameter_search_space']["dropout"] = tune.choice(config['hyperparameter_search_space']['dropout'])
    
    if config['combination']['model_label'] not in ['mlp', 'cnn']:
        config['hyperparameter_search_space']["num_layers"] = tune.choice(config['hyperparameter_search_space']['num_layers'])
    
    return config

# ----------------------------------------------------------------------------------------------------
# ---------------------------------- Auxiliary function for models -----------------------------------
# ----------------------------------------------------------------------------------------------------

def calc_maxpool_output(hidden_size, sequence_length):
    conv1_padding = 0
    conv1_dilation = 1
    conv1_kernel_size = 12
    conv1_stride = 1

    l_out = ((sequence_length + 2*conv1_padding - conv1_dilation*(conv1_kernel_size-1) - 1)/conv1_stride + 1)
    maxpool_padding = 0
    maxpool_dilation = 1
    maxpool_stride = 5
    maxpool_kernel_size = 12
    max_pool_output = int((l_out+2*maxpool_padding-maxpool_dilation*(maxpool_kernel_size-1)-1)/maxpool_stride+1)

    max_pool_output *= hidden_size
    
    return max_pool_output


# ----------------------------------------------------------------------------------------------------
# ------------------- The following functions were retrieved from repDNA package ---------------------
# ----------------------------------------------------------------------------------------------------

def ready_acc(k, phyche_index=None, all_property=False, extra_phyche_index=None):
    """Public function for get sequence_list and phyche_value.
    """
    if phyche_index is None:
        phyche_index = []
    if extra_phyche_index is None:
        extra_phyche_index = {}
    phyche_value = generate_phyche_value(k, phyche_index, all_property, extra_phyche_index)

    return phyche_value


def generate_phyche_value(k, phyche_index=None, all_property=False, extra_phyche_index=None):
    """Combine the user selected phyche_list, is_all_property and extra_phyche_index to a new standard phyche_value."""
    if phyche_index is None:
        phyche_index = []
    if extra_phyche_index is None:
        extra_phyche_index = {}

    diphyche_list = ['Base stacking', 'Protein induced deformability', 'B-DNA twist', 'Dinucleotide GC Content',
                     'A-philicity', 'Propeller twist', 'Duplex stability:(freeenergy)',
                     'Duplex tability(disruptenergy)', 'DNA denaturation', 'Bending stiffness', 'Protein DNA twist',
                     'Stabilising energy of Z-DNA', 'Aida_BA_transition', 'Breslauer_dG', 'Breslauer_dH',
                     'Breslauer_dS', 'Electron_interaction', 'Hartman_trans_free_energy', 'Helix-Coil_transition',
                     'Ivanov_BA_transition', 'Lisser_BZ_transition', 'Polar_interaction', 'SantaLucia_dG',
                     'SantaLucia_dH', 'SantaLucia_dS', 'Sarai_flexibility', 'Stability', 'Stacking_energy',
                     'Sugimoto_dG', 'Sugimoto_dH', 'Sugimoto_dS', 'Watson-Crick_interaction', 'Twist', 'Tilt',
                     'Roll', 'Shift', 'Slide', 'Rise']
    triphyche_list = ['Dnase I', 'Bendability (DNAse)', 'Bendability (consensus)', 'Trinucleotide GC Content',
                      'Nucleosome positioning', 'Consensus_roll', 'Consensus-Rigid', 'Dnase I-Rigid', 'MW-Daltons',
                      'MW-kg', 'Nucleosome', 'Nucleosome-Rigid']

    # Set and check physicochemical properties.
    if 2 == k:
        if all_property is True:
            phyche_index = diphyche_list
        else:
            for e in phyche_index:
                if e not in diphyche_list:
                    raise ValueError(" ".join(["Sorry, the physicochemical properties", e, "is not exit."]))
    elif 3 == k:
        if all_property is True:
            phyche_index = triphyche_list
        else:
            for e in phyche_index:
                if e not in triphyche_list:
                    raise ValueError(" ".join(["Sorry, the physicochemical properties", e, "is not exit."]))

    return extend_phyche_index(get_phyche_index(k, phyche_index), extra_phyche_index)


def get_phyche_index(k, phyche_list):
    """get phyche_value according phyche_list."""
    phyche_value = {}
    if 0 == len(phyche_list):
        for nucleotide in make_kmer_list(k):
            phyche_value[nucleotide] = []
        return phyche_value

    nucleotide_phyche_value = get_phyche_factor_dic(k)
    for nucleotide in make_kmer_list(k):
        if nucleotide not in phyche_value:
            phyche_value[nucleotide] = []
        for e in nucleotide_phyche_value[nucleotide]:
            if e[0] in phyche_list:
                phyche_value[nucleotide].append(e[1])

    return phyche_value


def extend_phyche_index(original_index, extend_index):
    """Extend {phyche:[value, ... ]}"""
    if extend_index is None or len(extend_index) == 0:
        return original_index
    for key in list(original_index.keys()):
        original_index[key].extend(extend_index[key])
    return original_index


def get_phyche_factor_dic(k):
    """Get all {nucleotide: [(phyche, value), ...]} dict."""
    full_path = os.path.realpath(__file__)
    if 2 == k:
        file_path = "%s/data/mmc3.data" % os.path.dirname(full_path)
    elif 3 == k:
        file_path = "%s/data/mmc4.data" % os.path.dirname(full_path)
    else:
        sys.stderr.write("The k can just be 2 or 3.")
        sys.exit(0)

    try:
        with open(file_path, 'rb') as f:
            phyche_factor_dic = pickle.load(f)
    except:
        with open(file_path, 'r') as f:
            phyche_factor_dic = pickle.load(f)

    return phyche_factor_dic


def make_ac_vector(sequence_list, lag, phyche_value, k):
    phyche_values = list(phyche_value.values())
    len_phyche_value = len(phyche_values[0])

    vec_ac = []
    for sequence in sequence_list:
        len_seq = len(sequence)
        each_vec = []

        for temp_lag in range(1, lag + 1):
            for j in range(len_phyche_value):

                # Calculate average phyche_value for a nucleotide.
                ave_phyche_value = 0.0
                for i in range(len_seq - temp_lag - k + 1):
                    nucleotide = sequence[i: i + k]
                    ave_phyche_value += float(phyche_value[nucleotide][j])
                ave_phyche_value /= len_seq

                # Calculate the vector.
                temp_sum = 0.0
                for i in range(len_seq - temp_lag - k + 1):
                    nucleotide1 = sequence[i: i + k]
                    nucleotide2 = sequence[i + temp_lag: i + temp_lag + k]
                    temp_sum += (float(phyche_value[nucleotide1][j]) - ave_phyche_value) * (
                        float(phyche_value[nucleotide2][j]))

                try:
                    val = round(temp_sum / (len_seq - temp_lag - k + 1), 3)
                except ZeroDivisionError:
                    val = 0.0
                each_vec.append(val)
        vec_ac.append(each_vec)

    return vec_ac


def make_cc_vector(sequence_list, lag, phyche_value, k):
    phyche_values = list(phyche_value.values())
    len_phyche_value = len(phyche_values[0])

    vec_cc = []
    for sequence in sequence_list:
        len_seq = len(sequence)
        each_vec = []

        for temp_lag in range(1, lag + 1):
            for i1 in range(len_phyche_value):
                for i2 in range(len_phyche_value):
                    if i1 != i2:
                        # Calculate average phyche_value for a nucleotide.
                        ave_phyche_value1 = 0.0
                        ave_phyche_value2 = 0.0
                        for j in range(len_seq - temp_lag - k + 1):
                            nucleotide = sequence[j: j + k]
                            ave_phyche_value1 += float(phyche_value[nucleotide][i1])
                            ave_phyche_value2 += float(phyche_value[nucleotide][i2])
                        ave_phyche_value1 /= len_seq
                        ave_phyche_value2 /= len_seq

                        # Calculate the vector.
                        temp_sum = 0.0
                        for j in range(len_seq - temp_lag - k + 1):
                            nucleotide1 = sequence[j: j + k]
                            nucleotide2 = sequence[j + temp_lag: j + temp_lag + k]
                            temp_sum += (float(phyche_value[nucleotide1][i1]) - ave_phyche_value1) * \
                                        (float(phyche_value[nucleotide2][i2]) - ave_phyche_value2)
                        try:
                            val = round(temp_sum / (len_seq - temp_lag - k + 1), 3)
                        except ZeroDivisionError:
                            val = 0.0
                        each_vec.append(val)

        vec_cc.append(each_vec)

    return vec_cc



"""
########################################################################
Runs a combination of hyperparameters or performs hyperparameter tuning
for the given model, feature mode, and data directory.
########################################################################
"""

import torch
import os
from .prepare_data import prepare_data
from .test import test
from .hyperparameter_tuning import hyperparameter_tuning
from .train import traindata
from .utils import print_metrics, read_config

os.environ["CUDA_VISIBLE_DEVICES"] = '1,2,3,4,5'
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

def perform(config):
    if config['do_tuning']:
        hyperparameter_tuning(device, config)
    else:
        model_label = config['combination']['model_label']
        mode = config['combination']['mode']
        data_dir = config['combination']['data_dir']
        class_weights = config['combination']['class_weights']
        batch_size = config['hyperparameters']['batch_size']
        kmer_one_hot = config['fixed_vals']['kmer_one_hot']
        hyperparameters = config['hyperparameters']
        
        trainloader, testloader, validloader, input_size, sequence_length = prepare_data(
            data_dir=data_dir,
            mode=mode,
            batch_size=batch_size,
            k=kmer_one_hot,
        )
        
        # train the model
        model = traindata(hyperparameters, device, config, trainloader, validloader, input_size, sequence_length)

        # test the model
        metrics = test(device, model, testloader)
        print_metrics(model_label, mode, data_dir, kmer_one_hot, class_weights, metrics)
        
if __name__ == '__main__':
    config = read_config(device)
    perform(config)


import torch
from torch import nn

import sys
sys.path.append("../")
from .utils import calc_maxpool_output

class MLP(nn.Module):
    """
    Implementation of DeepHe's MLP model
    - Paper: https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1008229
    - Code using Keras: https://github.com/xzhang2016/DeepHE/blob/master/DNN.py
    """

    def __init__(self, input_size, hidden_size, output_size, dropout):
        super(MLP, self).__init__()

        self.linear = nn.Linear(input_size, hidden_size * 2)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout)
        self.fc_last = nn.Linear(hidden_size * 2, output_size)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = self.linear(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.fc_last(x)
        x = self.sigmoid(x)
        return x


class CNN(nn.Module):
    """
    Implementation of Primer's CNN model (https://github.com/abidlabs/deep-learning-genomics-primer/blob/master/A_Primer_on_Deep_Learning_in_Genomics_Public.ipynb)
    """

    def __init__(self, input_size, hidden_size, output_size, dropout, sequence_length):
        super(CNN, self).__init__()

        self.conv1 = nn.Conv1d(input_size, hidden_size, kernel_size=12, stride=1, padding=0, dilation=1)
        self.maxpool = nn.MaxPool1d(kernel_size=12, stride=5, padding=0, dilation=1)
        
        max_pool_output = calc_maxpool_output(hidden_size, sequence_length)
        
        self.linear = nn.Linear(max_pool_output, hidden_size * 2)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout)

        self.fc_last = nn.Linear(hidden_size * 2, output_size)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        x = x.permute(0, 2, 1)
        x = self.conv1(x)
        x = self.maxpool(x)
        x = torch.flatten(x, 1)
        x = self.linear(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.fc_last(x)
        x = self.softmax(x)
        return x
    
class LSTM(nn.Module):
    """
    https://github.com/aladdinpersson/Machine-Learning-Collection/blob/master/ML/Pytorch/Basics/pytorch_rnn_gru_lstm.py
    """

    def __init__(self, input_size, hidden_size, is_bidirectional, num_layers, output_size, sequence_length, dropout, device):
        super(LSTM, self).__init__()
        self.num_directions = 2 if is_bidirectional else 1
        self.hidden_size = hidden_size
        self.device = device
        self.num_layers = num_layers
        
        self.lstm = nn.LSTM(input_size, hidden_size, num_layers, batch_first=True, bidirectional=is_bidirectional, dropout=dropout)
        self.fc = nn.Linear(hidden_size * sequence_length * self.num_directions, output_size)

    def forward(self, x):
        h0 = torch.zeros(self.num_layers * self.num_directions, x.size(0), self.hidden_size).to(self.device)
        c0 = torch.zeros(self.num_layers * self.num_directions, x.size(0), self.hidden_size).to(self.device)

        out, _ = self.lstm(x, (h0, c0))  
        out = out.reshape(out.shape[0], -1)
        out = self.fc(out)
        return out

class GRU(nn.Module):
    """
    https://github.com/aladdinpersson/Machine-Learning-Collection/blob/master/ML/Pytorch/Basics/pytorch_rnn_gru_lstm.py
    """

    def __init__(self, input_size, hidden_size, is_bidirectional, num_layers, output_size, sequence_length, dropout, device):
        super(GRU, self).__init__()
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.device = device
        self.num_directions = 2 if is_bidirectional else 1

        self.gru = nn.GRU(input_size, hidden_size, num_layers, batch_first=True, bidirectional=is_bidirectional, dropout=dropout)
        self.fc = nn.Linear(hidden_size * sequence_length * self.num_directions, output_size)

    def forward(self, x):
        h0 = torch.zeros(self.num_layers * self.num_directions, x.size(0), self.hidden_size).to(self.device)

        out, _ = self.gru(x, h0)
        out = out.reshape(out.shape[0], -1)
        out = self.fc(out)
        return out

class CNN_LSTM(nn.Module):
    
    def __init__(self, input_size, hidden_size, is_bidirectional, num_layers, sequence_length, output_size, dropout, device):
        super(CNN_LSTM, self).__init__()
        self.num_directions = 2 if is_bidirectional else 1
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.device = device

        self.conv1 = nn.Conv1d(input_size, hidden_size, kernel_size=12, stride=1, padding=0, dilation=1)
        self.maxpool = nn.MaxPool1d(kernel_size=12, stride=5, padding=0, dilation=1)
        
        max_pool_output = calc_maxpool_output(hidden_size, sequence_length)
        
        self.linear = nn.Linear(max_pool_output, hidden_size * 2)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout)
        
        self.lstm = nn.LSTM(hidden_size * 2, hidden_size, num_layers, batch_first=True, bidirectional=is_bidirectional, dropout=dropout)
        self.last_linear = nn.Linear(hidden_size * self.num_directions, output_size)

    def forward(self, x):
        h0 = torch.zeros(self.num_layers * self.num_directions, x.size(0), self.hidden_size).to(self.device)
        c0 = torch.zeros(self.num_layers * self.num_directions, x.size(0), self.hidden_size).to(self.device)

        x = x.permute(0, 2, 1)
        x = self.conv1(x)
        x = self.maxpool(x)
        x = torch.flatten(x, 1)
        x = self.linear(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = x.reshape(x.size(0), 1, -1) 
        out, _ = self.lstm(x, (h0, c0))
        out = out.reshape(out.shape[0], -1)
        y = self.last_linear(out)
        return y
    
class CNN_GRU(nn.Module):
    """
    https://medium.com/geekculture/recap-of-how-to-implement-lstm-in-pytorch-e17ec11b061e
    """
    def __init__(self, input_size, hidden_size, is_bidirectional, num_layers, sequence_length, output_size, dropout, device):
        super(CNN_GRU, self).__init__()
        
        self.num_directions = 2 if is_bidirectional else 1
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.device = device
        
        self.conv1 = nn.Conv1d(input_size, hidden_size, kernel_size=12, stride=1, padding=0, dilation=1)
        self.maxpool = nn.MaxPool1d(kernel_size=12, stride=5, padding=0, dilation=1)
        
        max_pool_output = calc_maxpool_output(hidden_size, sequence_length)
        
        self.linear = nn.Linear(max_pool_output, hidden_size * 2)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout)
        
        self.gru = nn.GRU(hidden_size * 2, hidden_size, num_layers, batch_first=True, bidirectional=is_bidirectional, dropout=dropout)
        self.last_linear = nn.Linear(hidden_size * self.num_directions, output_size)

    def forward(self, x):
        h0 = torch.zeros(self.num_layers * self.num_directions, x.size(0), self.hidden_size).to(self.device)

        x = x.permute(0, 2, 1)
        x = self.conv1(x)
        x = self.maxpool(x)
        x = torch.flatten(x, 1)
        x = self.linear(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = x.reshape(x.size(0), 1, -1)     
        out, _ = self.gru(x, h0)
        out = out.reshape(out.shape[0], -1)
        y = self.last_linear(out)
        return y
    


# -*- coding: utf-8 -*-
"""
##############################################################################

A class used for reading sequences or change sequences.
The main objective is to create sequence objects to calculate descriptors
The class allows to:
     1)Read sequences from string or from uniprot ID (is also possible retrieve sequences fom txt with uniprot IDs)
     2)Check if the protein sequence is a valid sequence
     3)Obtain a sized sequence of list of sequences, adding or cutting from both n and c terminals
     4)From one sequence generate list of subsequences based on sliding window approach, from specific aa, from the
     the terminals or divide the sequence in parts

Authors:Ana Marta Sequeira

Date: 01/2019

Email:

##############################################################################
"""

from adjuv_functions.sequence.get_sequence import get_protein_sequence, get_protein_sequence_from_txt
from adjuv_functions.sequence.pro_check import protein_check
from adjuv_functions.sequence.get_sized_seq import seq_equal_lenght
from adjuv_functions.sequence.get_sub_seq import sub_seq_sliding_window, sub_seq_to_aa, sub_seq_split,sub_seq_terminals


class ReadSequence:

    """
    The ReadSequence class aims to read the input and transform it into a sequence that can be used to calculate Descriptors.
    It accepts uniprotID, txt with uniprot ID and string with aminoacid sequence. The functions to read protein are based on the package pydpi.
    It allows to get sized sequences or a variety of subsequences from just one sequence.
    """

    def __init__(self):
        """	constructor """
        self.protein_sequence = None

        #check the best way todo this
        self.alphabet = "XARNDCEQGHILKMFPSTWYV"
    # alphabet = "ARNDCEQGHILKMFPSTWYV"
    # alphabet_x = "XARNDCEQGHILKMFPSTWYV"
    # alphabet_all_characters = "XARNDCEQGHILKMFPSTWYVBZUO"

# #########################################
# GET
    def get_seq(self):
        return self.protein_sequence

# #########################################
    # read string or ID and get protein sequence

    def read_protein_sequence(self, protein_sequence=""):
        """
        Read a protein sequence.
        :param protein_sequence: String with sequence
        :return: String with aa sequence
        """

        self.protein_sequence=str.strip(protein_sequence)
        # index=protein_check(protein_sequence)

        # if index==0:
        #     print("Error......")
        #     print("Please input a correct protein.")
        # else:
        return self.protein_sequence

    def get_protein_sequence_from_txt(self, path, openfile, savefile):
        """
        Function to retrieve sequences from a txt with uniprot ID. Does not retrieve a sequence object
        :param path: directory path containing ID file
        :param openfile:ID file ('name.txt')
        :param savefile: saved file with obtained protein sequences ('name2.txt')
        :return:File containing string of sequences
        """
        self.txt=get_protein_sequence_from_txt(path,openfile,savefile)
        return self.txt

    def get_protein_sequence_from_id(self, uniprotid=""):
        """
        Downloading a protein sequence by uniprot id.
        :param uniprotid: String with UniprotID
        :return: String with aa sequence
        """
        self.protein=get_protein_sequence(str.strip(uniprotid))
        index=protein_check(self.protein)

        if index==0:
            print("Error......")
            print("Please input a correct protein.")
        else:
            self.protein_sequence=self.protein
        return self.protein_sequence

##########################################
    # check protein

    def checkprotein(self, protein_sequence=""):
        """
        Check whether the protein sequence is a valid amino acid sequence or not.
        Just check. Not assign to the object.
        :param protein_sequence: protein sequence
        :return: error or valid message.
        """

        index=protein_check(protein_sequence)

        if index==0:
            print("Error......")
            print("Please input a correct protein.")
        else:
            print('sequence valid')

##########################################
# deal with strange AA
    #     alphabet = "ARNDCEQGHILKMFPSTWYV"
    # alphabet_x = "XARNDCEQGHILKMFPSTWYV"
    # alphabet_all_characters = "XARNDCEQGHILKMFPSTWYVBZUO"
    # TODO check the best way to do this

    def deal_with_strange_aa(self, seq=None, alphabet = "XARNDCEQGHILKMFPSTWYV",
                             B='N', Z = 'Q', U = 'C', O = 'K', X = '' ):
        """

        :param seq:
        :param alphabet:
        :param B:
        :param Z:
        :param U:
        :param O:
        :param X:
        :return:
        """
        if seq is None:
            seq = self.protein_sequence

        if len(alphabet) < 25:  # alphabet x or alphabet normal
            seq1 = seq.replace('B', B)  # asparagine N / aspartic acid  D - asx - B
            seq2 = seq1.replace('Z', Z)  # glutamine Q / glutamic acid  E - glx - Z
            seq3 = seq2.replace('U', U)  # selenocisteina, the closest is the cisteine. but it is a different aminoacid . take care.
            seq4 = seq3.replace('O', O)  # Pyrrolysine to lysine
            seq = seq4
            if len(alphabet) == 20:  # alphabet normal substitute every letters
                seq = seq4.replace('X', X)  # unknown character eliminated
        self.protein_sequence = seq
        return seq

##########################################
    # Generate subsequences

    def get_sub_seq_sliding_window(self, seq, window_size=20, gap=1, index=True):
        """
        sliding window of the protein given. It will generate a list of n sequences with
        lenght equal to the value of window and spaced a gap value. It can or not retrieve
        the indices of location of the subsequence in the original sequence.

        :param seq: protein sequence
        :param window_size: number of aminoacids to considerer, lenght of the subsequence. for default 20
        :param gap: gap size of the search of windows in sequence. default 1
        :param index: if true, return the indices of location of the subsequence in the original sequence
        :return: list with subsequences generated with or without a list of tuples with location of subsequences
         in original sequence
        """

        list_sliding_window=sub_seq_sliding_window(seq,window_size,gap,index)
        return list_sliding_window

    def get_sub_seq_to_aa(self, seq, ToAA='S', window=5):
        """
        Get all 2*window+1 sub-sequences whose center is ToAA in a protein

        :param seq:  protein sequence
        :param ToAA: central (query point) amino acid in the sub-sequence
        :param window:  span (number of amnoacids to go front and back from the ToAA
        :return: list form containing sub-sequences
        """

        list_subseq_to_aa=sub_seq_to_aa(seq, ToAA, window)
        return list_subseq_to_aa

    def get_sub_seq_split(self, seq, number_of_subseq):
        """
        Split the original seq in n number of subsequences.
        :param seq: protein sequence
        :param number_of_subseq: number of subsequences to divide the original seq
        :return: list with number_of_sequences sequences.

        """
        list_sub_seq_split=sub_seq_split(seq,number_of_subseq)
        return list_sub_seq_split

    ##########
    ### TODO change this

    def get_sub_seq_terminals(self, seq, N_terminal=5, C_terminal=5, rest=True):
        """
        Divide the sequence in the N terminal and  C terminal with sizes defined by the user. It returns
        a list with N and C terminal and rest of the sequence.
        By default the N terminal is considerer to be the beggining of the sequence (left)

        :param seq: protein sequence
        :param N_terminal: size of the N terminal to consider. If zero will not return
        :param C_terminal: size of the C terminal to consider. If zero will not return
        :param rest: If true will return the restant subsequence
        :return: list with N, C and rest of the sequence
        """
        list_sub_seq_term= sub_seq_terminals(seq, N_terminal, C_terminal,rest)
        return list_sub_seq_term


    ##########################################
    # Get equal size sequences is with a dataset no with sequence

    # def get_terminals(seq, l):
    #     if len(seq)<=l:
    #         return seq
    #     else:
    #         le=int(l/2)
    #         term_c=seq[:le]
    #         term_n=seq[-le:]
    #         terminals=term_c+term_n
    #         return terminals
    #
    # def get_middle(seq,l):
    #     if int(len(seq))<=l:
    #         return seq
    #     else:
    #         takeoff = int(len(seq)-l)
    #         if (len(seq) % 2) == 0:
    #             takeoff_2 = takeoff//2
    #             term_c = takeoff_2
    #             term_n = takeoff_2
    #         else:
    #             takeoff_2 = takeoff//2
    #             term_c = takeoff_2
    #             term_n = takeoff_2 + 1
    #         middle = seq[term_c:-term_n]
    #         return middle


    def get_sized_seq(self, sequences=None, n_terminal=10, c_terminal=10, terminal=0, dummie = 'X'):
        # todo put here pad sequences
        # todo check the part of enZyme classification file

        """
        cut or add aminoacids to obtain sequences with equal lenght.
        :param sequences: list containing protein sequences (string) or just a protein sequence
        :param n_terminal: number of aa to consider in the n terminal (left side of sequence)
        :param c_terminal: number of aa to consider in the c terminal (right side of sequence)
        :param terminal: in case of the need to add dummy aa and no terminal as already been chosen, it decides where to add
                        0 to add to the right (consider N terminal)
                        -1 to add to the left (consider C terminal)
                        2 to add in the middle (N and C terminal will be both present and repeated with dummy in middle
        :return: list of sequences containing all the same lenght. if just one sequence given it will return a string
        """
        if len(sequences)==1:
            seq=seq_equal_lenght(sequences[0],n_terminal,c_terminal,terminal, dummie=dummie)
            return seq
        else:
            equal_size_sequences=[] #list to store sequences with equal lenght
            for seq in sequences:
                seq_2=seq_equal_lenght(seq,n_terminal,c_terminal,terminal, dummie=dummie)
                equal_size_sequences.append(seq_2)
            return equal_size_sequences
